# cython: language_level=2
cimport cUnitSquareImage
import numpy as np

cimport numpy as np
from cpython cimport array

from libcpp cimport bool

np.import_array()

def compgrad3D(np.ndarray[double, ndim=3, mode="c"] F):
    """
    cython interface calculates the 2D gradient of q-map image

    :param F: matrix of three dimension 

    :rtype numpy ndarray
    :return dfdu: derivative
    :return dfdv: derivative

    """
    cdef int n, t, d
    cdef double lam

    n = F.shape[0]
    t = F.shape[1]
    d = F.shape[2]

    cdef np.ndarray[double, ndim=1, mode="c"] Fi = np.zeros(n*t*d)
    cdef np.ndarray[double, ndim=1, mode="c"] dfdui = np.zeros(n*t*d)
    cdef np.ndarray[double, ndim=1, mode="c"] dfdvi = np.zeros(n*t*d)

    Fi = F.reshape(n*t*d, order='F')

    Fi = np.ascontiguousarray(Fi)

    cUnitSquareImage.findgrad2D(&dfdui[0], &dfdvi[0], &Fi[0], n, t, d)

    return dfdui, dfdvi


def compgrad2D(np.ndarray[double, ndim=2, mode="c"] F):
    """
    cython interface calculates the 2D gradient of q-map image

    :param F: matrix of three dimension 

    :rtype numpy ndarray
    :return dfdu: derivative
    :return dfdv: derivative

    """
    cdef int n, t, d
    cdef double lam

    n = F.shape[0]
    t = F.shape[1]
    d = 1

    cdef np.ndarray[double, ndim=1, mode="c"] Fi = np.zeros(n*t*d)
    cdef np.ndarray[double, ndim=1, mode="c"] dfdui = np.zeros(n*t*d)
    cdef np.ndarray[double, ndim=1, mode="c"] dfdvi = np.zeros(n*t*d)

    Fi = F.reshape(n*t*d, order='F')

    Fi = np.ascontiguousarray(Fi)

    cUnitSquareImage.findgrad2D(&dfdui[0], &dfdvi[0], &Fi[0], n, t, d)

    return dfdui, dfdvi


def check_crossing(np.ndarray[double, ndim=3, mode="c"] gam):
    cdef int n, t, D, is_diffeo

    n = gam.shape[0]
    t = gam.shape[1]
    D = gam.shape[2]

    is_diffeo = 1

    cdef np.ndarray[double, ndim=1, mode="c"] gami = np.zeros(n*t*D)

    gami = gam.reshape(n*t*D, order='F')

    gami = np.ascontiguousarray(gami)

    is_diffeo = cUnitSquareImage.check_crossing(&gami[0], n, t, D)

    return is_diffeo



"""
Warping Invariant Regression using SRSF

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

import numpy as np
import fdasrsf.utility_functions as uf
from scipy.optimize import fmin_l_bfgs_b
from scipy.integrate import trapezoid
from scipy.linalg import inv, norm
from patsy import bs
from joblib import Parallel, delayed
import mlogit_warp as mw


class elastic_regression:
    """
    This class provides elastic regression for functional data using the
    SRVF framework accounting for warping

    Usage:  obj = elastic_regression(f,y,time)

    :param f: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy array of N responses
    :param time: vector of size M describing the sample points
    :param B: optional matrix describing Basis elements
    :param alpha: alpha parameter of model
    :param beta: beta(t) of model
    :param fn: aligned functions - numpy ndarray of shape (M,N) of M functions with N samples
    :param qn: aligned srvfs - similar structure to fn
    :param gamma: calculated warping functions
    :param q: original training SRSFs
    :param b: basis coefficients
    :param SSE: sum of squared error

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  29-Oct-2021
    """

    def __init__(self, f, y, time):
        """
        Construct an instance of the elastic_regression class
        :param f: numpy ndarray of shape (M,N) of N functions with M samples
        :param y: response vector
        :param time: vector of size M describing the sample points
        """
        a = time.shape[0]

        if f.shape[0] != a:
            raise Exception("Columns of f and time must be equal")

        self.f = f
        self.y = y
        self.time = time

    def calc_model(self, B=None, lam=0, df=20, max_itr=20, cores=-1, smooth=False):
        """
        This function identifies a regression model with phase-variability
        using elastic pca

        :param B: optional matrix describing Basis elements
        :param lam: regularization parameter (default 0)
        :param df: number of degrees of freedom B-spline (default 20)
        :param max_itr: maximum number of iterations (default 20)
        :param cores: number of cores for parallel processing (default all)
        """

        M = self.f.shape[0]
        N = self.f.shape[1]

        if M > 500:
            parallel = True
        elif N > 100:
            parallel = True
        else:
            parallel = False

        binsize = np.diff(self.time)
        binsize = binsize.mean()

        # Create B-Spline Basis if none provided
        if B is None:
            B = bs(self.time, df=df, degree=4, include_intercept=True)
        Nb = B.shape[1]

        self.B = B

        # second derivative for regularization
        Bdiff = np.zeros((M, Nb))
        for ii in range(0, Nb):
            Bdiff[:, ii] = np.gradient(np.gradient(B[:, ii], binsize), binsize)

        self.Bdiff = Bdiff

        self.q = uf.f_to_srsf(self.f, self.time, smooth)

        gamma = np.tile(np.linspace(0, 1, M), (N, 1))
        gamma = gamma.transpose()

        itr = 1
        self.SSE = np.zeros(max_itr)
        while itr <= max_itr:
            print("Iteration: %d" % itr)
            # align data
            fn = np.zeros((M, N))
            qn = np.zeros((M, N))
            for ii in range(0, N):
                fn[:, ii] = np.interp(
                    (self.time[-1] - self.time[0]) * gamma[:, ii] + self.time[0],
                    self.time,
                    self.f[:, ii],
                )
                qn[:, ii] = uf.warp_q_gamma(self.time, self.q[:, ii], gamma[:, ii])

            # OLS using basis
            Phi = np.ones((N, Nb + 1))
            for ii in range(0, N):
                for jj in range(1, Nb + 1):
                    Phi[ii, jj] = trapezoid(qn[:, ii] * B[:, jj - 1], self.time)

            R = np.zeros((Nb + 1, Nb + 1))
            for ii in range(1, Nb + 1):
                for jj in range(1, Nb + 1):
                    R[ii, jj] = trapezoid(Bdiff[:, ii - 1] * Bdiff[:, jj - 1], self.time)

            xx = np.dot(Phi.T, Phi)
            inv_xx = inv(xx + lam * R)
            xy = np.dot(Phi.T, self.y)
            b = np.dot(inv_xx, xy)

            alpha = b[0]
            beta = B.dot(b[1 : Nb + 1])
            beta = beta.reshape(M)

            # compute the SSE
            int_X = np.zeros(N)
            for ii in range(0, N):
                int_X[ii] = trapezoid(qn[:, ii] * beta, self.time)

            self.SSE[itr - 1] = sum((self.y.reshape(N) - alpha - int_X) ** 2)

            # find gamma
            gamma_new = np.zeros((M, N))
            if parallel:
                out = Parallel(n_jobs=cores)(
                    delayed(regression_warp)(
                        beta, self.time, self.q[:, n], self.y[n], alpha
                    )
                    for n in range(N)
                )
                gamma_new = np.array(out)
                gamma_new = gamma_new.transpose()
            else:
                for ii in range(0, N):
                    gamma_new[:, ii] = regression_warp(
                        beta, self.time, self.q[:, ii], self.y[ii], alpha
                    )

            if norm(gamma - gamma_new) < 1e-5:
                break
            else:
                gamma = gamma_new

            itr += 1

        # Last Step with centering of gam
        gamI = uf.SqrtMeanInverse(gamma_new)
        gamI_dev = np.gradient(gamI, 1 / float(M - 1))
        beta = np.interp(
            (self.time[-1] - self.time[0]) * gamI + self.time[0], self.time, beta
        ) * np.sqrt(gamI_dev)

        for ii in range(0, N):
            qn[:, ii] = np.interp(
                (self.time[-1] - self.time[0]) * gamI + self.time[0],
                self.time,
                qn[:, ii],
            ) * np.sqrt(gamI_dev)
            fn[:, ii] = np.interp(
                (self.time[-1] - self.time[0]) * gamI + self.time[0],
                self.time,
                fn[:, ii],
            )
            gamma[:, ii] = np.interp(
                (self.time[-1] - self.time[0]) * gamI + self.time[0],
                self.time,
                gamma_new[:, ii],
            )

        self.qn = qn
        self.fn = fn
        self.gamma = gamma
        self.alpha = alpha
        self.beta = beta
        self.b = b[1:-1]
        self.SSE = self.SSE[0:itr]

        return

    def predict(self, newdata=None):
        """
        This function performs prediction on regression model on new data if available or current stored data in object
        Usage:  obj.predict(newdata)

        :param newdata: dict containing new data for prediction (needs the keys below, if None predicts on training data)
        :type newdata: dict
        :param f: (M,N) matrix of functions
        :param time: vector of time points
        :param y: truth if available
        :param smooth: smooth data if needed
        :param sparam: number of times to run filter
        """

        if newdata != None:
            f = newdata["f"]
            time = newdata["time"]
            y = newdata["y"]

            q = uf.f_to_srsf(f, time, newdata["smooth"])

            n = f.shape[1]
            yhat = np.zeros(n)
            for ii in range(0, n):
                diff = self.q - q[:, ii][:, np.newaxis]
                dist = np.sum(np.abs(diff) ** 2, axis=0) ** (1.0 / 2)
                q_tmp = uf.warp_q_gamma(time, q[:, ii], self.gamma[:, dist.argmin()])
                yhat[ii] = self.alpha + trapezoid(q_tmp * self.beta, time)

            if y is None:
                self.SSE = np.nan
            else:
                self.SSE = np.sum((y - yhat) ** 2)

            self.y_pred = yhat

        else:
            n = self.f.shape[1]
            yhat = np.zeros(n)
            for ii in range(0, n):
                diff = self.q - self.q[:, ii][:, np.newaxis]
                dist = np.sum(np.abs(diff) ** 2, axis=0) ** (1.0 / 2)
                q_tmp = uf.warp_q_gamma(
                    self.time, self.q[:, ii], self.gamma[:, dist.argmin()]
                )
                yhat[ii] = self.alpha + trapezoid(q_tmp * self.beta, self.time)

            self.SSE = np.sum((self.y - yhat) ** 2)
            self.y_pred = yhat

        return


class elastic_logistic:
    """
    This class provides elastic logistic regression for functional data using the
    SRVF framework accounting for warping

    Usage:  obj = elastic_logistic(f,y,time)

    :param f: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy array of N responses
    :param time: vector of size M describing the sample points
    :param B: optional matrix describing Basis elements
    :param alpha: alpha parameter of model
    :param beta: beta(t) of model
    :param fn: aligned functions - numpy ndarray of shape (M,N) of M functions with N samples
    :param qn: aligned srvfs - similar structure to fn
    :param gamma: calculated warping functions
    :param q: original training SRSFs
    :param b: basis coefficients
    :param Loss: logistic loss
    :type f: np.ndarray
    :type time: np.ndarray

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  29-Oct-2021
    """

    def __init__(self, f, y, time):
        """
        Construct an instance of the elastic_regression class
        :param f: numpy ndarray of shape (M,N) of N functions with M samples
        :param y: response vector
        :param time: vector of size M describing the sample points
        """
        a = time.shape[0]

        if f.shape[0] != a:
            raise Exception("Columns of f and time must be equal")

        self.f = f
        self.y = y
        self.time = time

    def calc_model(self, B=None, lam=0, df=20, max_itr=20, cores=-1, smooth=False):
        """
        This function identifies a regression model with phase-variability
        using elastic pca

        :param B: optional matrix describing Basis elements
        :param lam: regularization parameter (default 0)
        :param df: number of degrees of freedom B-spline (default 20)
        :param max_itr: maximum number of iterations (default 20)
        :param cores: number of cores for parallel processing (default all)
        """

        M = self.f.shape[0]
        N = self.f.shape[1]

        if M > 500:
            parallel = True
        elif N > 100:
            parallel = True
        else:
            parallel = False

        binsize = np.diff(self.time)
        binsize = binsize.mean()

        # Create B-Spline Basis if none provided
        if B is None:
            B = bs(self.time, df=df, degree=4, include_intercept=True)
        Nb = B.shape[1]

        self.B = B

        self.q = uf.f_to_srsf(self.f, self.time, smooth)

        gamma = np.tile(np.linspace(0, 1, M), (N, 1))
        gamma = gamma.transpose()

        itr = 1
        self.LL = np.zeros(max_itr)
        while itr <= max_itr:
            print("Iteration: %d" % itr)
            # align data
            fn = np.zeros((M, N))
            qn = np.zeros((M, N))
            for ii in range(0, N):
                fn[:, ii] = np.interp(
                    (self.time[-1] - self.time[0]) * gamma[:, ii] + self.time[0],
                    self.time,
                    self.f[:, ii],
                )
                qn[:, ii] = uf.warp_q_gamma(self.time, self.q[:, ii], gamma[:, ii])

            Phi = np.ones((N, Nb + 1))
            for ii in range(0, N):
                for jj in range(1, Nb + 1):
                    Phi[ii, jj] = trapezoid(qn[:, ii] * B[:, jj - 1], self.time)

            # Find alpha and beta using l_bfgs
            b0 = np.zeros(Nb + 1)
            out = fmin_l_bfgs_b(
                logit_loss,
                b0,
                fprime=logit_gradient,
                args=(Phi, self.y),
                pgtol=1e-10,
                maxiter=200,
                maxfun=250,
                factr=1e-30,
            )
            b = out[0]
            alpha = b[0]
            beta = B.dot(b[1 : Nb + 1])
            beta = beta.reshape(M)

            # compute the logistic loss
            self.LL[itr - 1] = logit_loss(b, Phi, self.y)

            # find gamma
            gamma_new = np.zeros((M, N))
            if parallel:
                out = Parallel(n_jobs=cores)(
                    delayed(logistic_warp)(beta, self.time, self.q[:, n], self.y[n])
                    for n in range(N)
                )
                gamma_new = np.array(out)
                gamma_new = gamma_new.transpose()
            else:
                for ii in range(0, N):
                    gamma_new[:, ii] = logistic_warp(
                        beta, self.time, self.q[:, ii], self.y[ii]
                    )

            if norm(gamma - gamma_new) < 1e-5:
                break
            else:
                gamma = gamma_new

            itr += 1

        self.qn = qn
        self.fn = fn
        self.gamma = gamma
        self.alpha = alpha
        self.beta = beta
        self.b = b[1:-1]
        self.LL = self.LL[0:itr]

        return

    def predict(self, newdata=None):
        """
        This function performs prediction on regression model on new data if available or current stored data in object
        Usage:  obj.predict(newdata)

        :param newdata: dict containing new data for prediction (needs the keys below, if None predicts on training data)
        :type newdata: dict
        :param f: (M,N) matrix of functions
        :param time: vector of time points
        :param y: truth if available
        :param smooth: smooth data if needed
        :param sparam: number of times to run filter
        """

        if newdata != None:
            f = newdata["f"]
            time = newdata["time"]
            y = newdata["y"]

            q = uf.f_to_srsf(f, time, newdata["smooth"])

            n = f.shape[1]
            yhat = np.zeros(n)
            for ii in range(0, n):
                diff = self.q - q[:, ii][:, np.newaxis]
                dist = np.sum(np.abs(diff) ** 2, axis=0) ** (1.0 / 2)
                q_tmp = uf.warp_q_gamma(time, q[:, ii], self.gamma[:, dist.argmin()])
                yhat[ii] = self.alpha + trapezoid(q_tmp * self.beta, time)

            if y is None:
                yhat = phi(yhat)
                y_labels = np.ones(n)
                y_labels[yhat < 0.5] = -1
                self.PC = None
            else:
                yhat = phi(yhat)
                y_labels = np.ones(n)
                y_labels[yhat < 0.5] = -1
                TP = sum(y[y_labels == 1] == 1)
                FP = sum(y[y_labels == -1] == 1)
                TN = sum(y[y_labels == -1] == -1)
                FN = sum(y[y_labels == 1] == -1)
                self.PC = (TP + TN) / float(TP + FP + FN + TN)

            self.y_pred = yhat
            self.y_labels = y_labels

        else:
            n = self.f.shape[1]
            yhat = np.zeros(n)
            for ii in range(0, n):
                diff = self.q - self.q[:, ii][:, np.newaxis]
                dist = np.sum(np.abs(diff) ** 2, axis=0) ** (1.0 / 2)
                q_tmp = uf.warp_q_gamma(
                    self.time, self.q[:, ii], self.gamma[:, dist.argmin()]
                )
                yhat[ii] = self.alpha + trapezoid(q_tmp * self.beta, self.time)

            yhat = phi(yhat)
            y_labels = np.ones(n)
            y_labels[yhat < 0.5] = -1
            TP = sum(self.y[y_labels == 1] == 1)
            FP = sum(self.y[y_labels == -1] == 1)
            TN = sum(self.y[y_labels == -1] == -1)
            FN = sum(self.y[y_labels == 1] == -1)
            self.PC = (TP + TN) / float(TP + FP + FN + TN)
            self.y_pred = yhat
            self.y_labels = y_labels

        return


class elastic_mlogistic:
    """
    This class provides elastic multinomial logistic regression for functional data using the
    SRVF framework accounting for warping

    Usage:  obj = elastic_mlogistic(f,y,time)

    :param f: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy array of N responses
    :param time: vector of size M describing the sample points
    :param B: optional matrix describing Basis elements
    :param alpha: alpha parameter of model
    :param beta: beta(t) of model
    :param fn: aligned functions - numpy ndarray of shape (M,N) of N functions with M samples
    :param qn: aligned srvfs - similar structure to fn
    :param gamma: calculated warping functions
    :param q: original training SRSFs
    :param b: basis coefficients
    :param Loss: logistic loss
    :type f: np.ndarray
    :type time: np.ndarray

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  29-Oct-2021
    """

    def __init__(self, f, y, time):
        """
        Construct an instance of the elastic_regression class
        :param f: numpy ndarray of shape (M,N) of N functions with M samples
        :param y: response vector
        :param time: vector of size M describing the sample points
        """
        a = time.shape[0]
        M = f.shape[0]
        N = f.shape[1]

        if f.shape[0] != a:
            raise Exception("Columns of f and time must be equal")

        self.f = f
        self.y = y

        # Code labels
        m = y.max()
        Y = np.zeros((N, m), dtype=int)
        for ii in range(0, N):
            Y[ii, y[ii] - 1] = 1
        self.Y = Y
        self.time = time

    def calc_model(
        self, B=None, lam=0, df=20, max_itr=20, delta=0.01, cores=-1, smooth=False
    ):
        """
        This function identifies a regression model with phase-variability
        using elastic pca

        :param B: optional matrix describing Basis elements
        :param lam: regularization parameter (default 0)
        :param df: number of degrees of freedom B-spline (default 20)
        :param max_itr: maximum number of iterations (default 20)
        :param cores: number of cores for parallel processing (default all)
        """

        M = self.f.shape[0]
        N = self.f.shape[1]
        m = self.y.max()

        if M > 500:
            parallel = True
        elif N > 100:
            parallel = True
        else:
            parallel = False

        binsize = np.diff(self.time)
        binsize = binsize.mean()

        # Create B-Spline Basis if none provided
        if B is None:
            B = bs(self.time, df=df, degree=4, include_intercept=True)
        Nb = B.shape[1]

        self.B = B

        self.q = uf.f_to_srsf(self.f, self.time, smooth)

        gamma = np.tile(np.linspace(0, 1, M), (N, 1))
        gamma = gamma.transpose()

        itr = 1
        self.LL = np.zeros(max_itr)
        while itr <= max_itr:
            print("Iteration: %d" % itr)
            # align data
            fn = np.zeros((M, N))
            qn = np.zeros((M, N))
            for ii in range(0, N):
                fn[:, ii] = np.interp(
                    (self.time[-1] - self.time[0]) * gamma[:, ii] + self.time[0],
                    self.time,
                    self.f[:, ii],
                )
                qn[:, ii] = uf.warp_q_gamma(self.time, self.q[:, ii], gamma[:, ii])

            Phi = np.ones((N, Nb + 1))
            for ii in range(0, N):
                for jj in range(1, Nb + 1):
                    Phi[ii, jj] = trapezoid(qn[:, ii] * B[:, jj - 1], self.time)

            # Find alpha and beta using l_bfgs
            b0 = np.zeros(m * (Nb + 1))
            out = fmin_l_bfgs_b(
                mlogit_loss,
                b0,
                fprime=mlogit_gradient,
                args=(Phi, self.Y),
                pgtol=1e-10,
                maxiter=200,
                maxfun=250,
                factr=1e-30,
            )
            b = out[0]
            B0 = b.reshape(Nb + 1, m)
            alpha = B0[0, :]
            beta = np.zeros((M, m))
            for i in range(0, m):
                beta[:, i] = B.dot(B0[1 : Nb + 1, i])

            # compute the logistic loss
            self.LL[itr - 1] = mlogit_loss(b, Phi, self.Y)

            # find gamma
            gamma_new = np.zeros((M, N))
            if parallel:
                out = Parallel(n_jobs=cores)(
                    delayed(mlogit_warp_grad)(
                        alpha, beta, self.time, self.q[:, n], self.Y[n, :], delta=delta
                    )
                    for n in range(N)
                )
                gamma_new = np.array(out)
                gamma_new = gamma_new.transpose()
            else:
                for ii in range(0, N):
                    gamma_new[:, ii] = mlogit_warp_grad(
                        alpha,
                        beta,
                        self.time,
                        self.q[:, ii],
                        self.Y[ii, :],
                        delta=delta,
                    )

            if norm(gamma - gamma_new) < 1e-5:
                break
            else:
                gamma = gamma_new

            itr += 1

        self.qn = qn
        self.fn = fn
        self.gamma = gamma
        self.alpha = alpha
        self.beta = beta
        self.b = b[1:-1]
        self.n_classes = m
        self.LL = self.LL[0:itr]

        return

    def predict(self, newdata=None):
        """
        This function performs prediction on regression model on new data if available or current stored data in object
        Usage:  obj.predict(newdata)

        :param newdata: dict containing new data for prediction (needs the keys below, if None predicts on training data)
        :type newdata: dict
        :param f: (M,N) matrix of functions
        :param time: vector of time points
        :param y: truth if available
        :param smooth: smooth data if needed
        :param sparam: number of times to run filter
        """

        if newdata != None:
            f = newdata["f"]
            time = newdata["time"]
            y = newdata["y"]

            q = uf.f_to_srsf(f, time, newdata["smooth"])

            n = f.shape[1]
            m = self.n_classes
            yhat = np.zeros((n, m))
            for ii in range(0, n):
                diff = self.q - q[:, ii][:, np.newaxis]
                dist = np.sum(np.abs(diff) ** 2, axis=0) ** (1.0 / 2)
                q_tmp = uf.warp_q_gamma(time, q[:, ii], self.gamma[:, dist.argmin()])
                for jj in range(0, m):
                    yhat[ii, jj] = self.alpha[jj] + trapezoid(
                        q_tmp * self.beta[:, jj], time
                    )

            if y is None:
                yhat = phi(yhat.ravel())
                yhat = yhat.reshape(n, m)
                y_labels = yhat.argmax(axis=1) + 1
                self.PC = None
            else:
                yhat = phi(yhat.ravel())
                yhat = yhat.reshape(n, m)
                y_labels = yhat.argmax(axis=1) + 1
                PC = np.zeros(m)
                cls_set = np.arange(1, m + 1)
                for ii in range(0, m):
                    cls_sub = np.delete(cls_set, ii)
                    TP = sum(y[y_labels == (ii + 1)] == (ii + 1))
                    FP = sum(y[np.in1d(y_labels, cls_sub)] == (ii + 1))
                    TN = sum(
                        y[np.in1d(y_labels, cls_sub)]
                        == y_labels[np.in1d(y_labels, cls_sub)]
                    )
                    FN = sum(np.in1d(y[y_labels == (ii + 1)], cls_sub))
                    PC[ii] = (TP + TN) / float(TP + FP + FN + TN)

                self.PC = sum(y == y_labels) / float(y_labels.size)

            self.y_pred = yhat
            self.y_labels = y_labels

        else:
            n = self.f.shape[1]
            m = self.n_classes
            yhat = np.zeros((n, m))
            for ii in range(0, n):
                diff = self.q - self.q[:, ii][:, np.newaxis]
                dist = np.sum(np.abs(diff) ** 2, axis=0) ** (1.0 / 2)
                q_tmp = uf.warp_q_gamma(
                    self.time, self.q[:, ii], self.gamma[:, dist.argmin()]
                )
                for jj in range(0, m):
                    yhat[ii, jj] = self.alpha[jj] + trapezoid(
                        q_tmp * self.beta[:, jj], self.time
                    )

            yhat = phi(yhat.ravel())
            yhat = yhat.reshape(n, m)
            y_labels = yhat.argmax(axis=1) + 1
            PC = np.zeros(m)
            cls_set = np.arange(1, m + 1)
            for ii in range(0, m):
                cls_sub = np.delete(cls_set, ii)
                TP = sum(self.y[y_labels == (ii + 1)] == (ii + 1))
                FP = sum(self.y[np.in1d(y_labels, cls_sub)] == (ii + 1))
                TN = sum(
                    self.y[np.in1d(y_labels, cls_sub)]
                    == y_labels[np.in1d(y_labels, cls_sub)]
                )
                FN = sum(np.in1d(self.y[y_labels == (ii + 1)], cls_sub))
                PC[ii] = (TP + TN) / float(TP + FP + FN + TN)

            self.PC = sum(self.y == y_labels) / float(y_labels.size)
            self.y_pred = yhat
            self.y_labels = y_labels

        return


# helper functions for linear regression
def regression_warp(beta, time, q, y, alpha):
    """
    calculates optimal warping for function linear regression

    :param beta: numpy ndarray of shape (M,N) of M functions with N samples
    :param time: vector of size N describing the sample points
    :param q: numpy ndarray of shape (M,N) of M functions with N samples
    :param y: numpy ndarray of shape (1,N) of M functions with N samples responses
    :param alpha: numpy scalar

    :rtype: numpy array
    :return gamma_new: warping function

    """
    gam_M = uf.optimum_reparam(beta, time, q)
    qM = uf.warp_q_gamma(time, q, gam_M)
    y_M = trapezoid(qM * beta, time)

    gam_m = uf.optimum_reparam(-1 * beta, time, q)
    qm = uf.warp_q_gamma(time, q, gam_m)
    y_m = trapezoid(qm * beta, time)

    if y > alpha + y_M:
        gamma_new = gam_M
    elif y < alpha + y_m:
        gamma_new = gam_m
    else:
        gamma_new = uf.zero_crossing(y - alpha, q, beta, time, y_M, y_m, gam_M, gam_m)

    return gamma_new


# helper functions for logistic regression
def logistic_warp(beta, time, q, y):
    """
    calculates optimal warping for function logistic regression

    :param beta: numpy ndarray of shape (M,N) of N functions with M samples
    :param time: vector of size N describing the sample points
    :param q: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy ndarray of shape (1,N) responses

    :rtype: numpy array
    :return gamma: warping function

    """
    if y == 1:
        gamma = uf.optimum_reparam(beta, time, q)
    elif y == -1:
        gamma = uf.optimum_reparam(-1 * beta, time, q)
    return gamma


def phi(t):
    """
    calculates logistic function, returns 1 / (1 + exp(-t))

    :param t: scalar

    :rtype: numpy array
    :return out: return value

    """
    # logistic function, returns 1 / (1 + exp(-t))
    idx = t > 0
    out = np.empty(t.size, dtype=np.float)
    out[idx] = 1.0 / (1 + np.exp(-t[idx]))
    exp_t = np.exp(t[~idx])
    out[~idx] = exp_t / (1.0 + exp_t)
    return out


def logit_loss(b, X, y):
    """
    logistic loss function, returns Sum{-log(phi(t))}

    :param b: numpy ndarray of shape (M,N) of N functions with M samples
    :param X: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy ndarray of shape (1,N) of N responses

    :rtype: numpy array
    :return out: loss value

    """
    z = X.dot(b)
    yz = y * z
    idx = yz > 0
    out = np.zeros_like(yz)
    out[idx] = np.log(1 + np.exp(-yz[idx]))
    out[~idx] = -yz[~idx] + np.log(1 + np.exp(yz[~idx]))
    out = out.sum()
    return out


def logit_gradient(b, X, y):
    """
    calculates gradient of the logistic loss

    :param b: numpy ndarray of shape (M,N) of N functions with M samples
    :param X: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy ndarray of shape (1,N) responses

    :rtype: numpy array
    :return grad: gradient of logistic loss

    """
    z = X.dot(b)
    z = phi(y * z)
    z0 = (z - 1) * y
    grad = X.T.dot(z0)
    return grad


def logit_hessian(s, b, X, y):
    """
    calculates hessian of the logistic loss

    :param s: numpy ndarray of shape (M,N) of N functions with M samples
    :param b: numpy ndarray of shape (M,N) of N functions with M samples
    :param X: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy ndarray of shape (1,N) responses

    :rtype: numpy array
    :return out: hessian of logistic loss

    """
    z = X.dot(b)
    z = phi(y * z)
    d = z * (1 - z)
    wa = d * X.dot(s)
    Hs = X.T.dot(wa)
    out = Hs
    return out


# helper functions for multinomial logistic regression
def mlogit_warp_grad(
    alpha, beta, time, q, y, max_itr=8000, tol=1e-10, delta=0.008, display=0
):
    """
    calculates optimal warping for functional multinomial logistic regression

    :param alpha: scalar
    :param beta: numpy ndarray of shape (M,N) of N functions with M samples
    :param time: vector of size M describing the sample points
    :param q: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy ndarray of shape (1,N) responses
    :param max_itr: maximum number of iterations (Default=8000)
    :param tol: stopping tolerance (Default=1e-10)
    :param delta: gradient step size (Default=0.008)
    :param display: display iterations (Default=0)

    :rtype: tuple of numpy array
    :return gam_old: warping function

    """

    gam_old = mw.mlogit_warp(
        np.ascontiguousarray(alpha),
        np.ascontiguousarray(beta),
        time,
        np.ascontiguousarray(q),
        np.ascontiguousarray(y, dtype=np.int32),
        max_itr,
        tol,
        delta,
        display,
    )

    return gam_old


def mlogit_loss(b, X, Y):
    """
    calculates multinomial logistic loss (negative log-likelihood)

    :param b: numpy ndarray of shape (M,N) of N functions with M samples
    :param X: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy ndarray of shape (1,N) responses

    :rtype: numpy array
    :return nll: negative log-likelihood

    """
    N, m = Y.shape  # n_samples, n_classes
    M = X.shape[1]  # n_features
    B = b.reshape(M, m)
    Yhat = np.dot(X, B)
    Yhat -= Yhat.min(axis=1)[:, np.newaxis]
    Yhat = np.exp(-Yhat)
    # l1-normalize
    Yhat /= Yhat.sum(axis=1)[:, np.newaxis]

    Yhat = Yhat * Y
    nll = np.sum(np.log(Yhat.sum(axis=1)))
    nll /= -float(N)

    return nll


def mlogit_gradient(b, X, Y):
    """
    calculates gradient of the multinomial logistic loss

    :param b: numpy ndarray of shape (M,N) of N functions with M samples
    :param X: numpy ndarray of shape (M,N) of N functions with M samples
    :param y: numpy ndarray of shape (1,N) responses

    :rtype: numpy array
    :return grad: gradient

    """
    N, m = Y.shape  # n_samples, n_classes
    M = X.shape[1]  # n_features
    B = b.reshape(M, m)
    Yhat = np.dot(X, B)
    Yhat -= Yhat.min(axis=1)[:, np.newaxis]
    Yhat = np.exp(-Yhat)
    # l1-normalize
    Yhat /= Yhat.sum(axis=1)[:, np.newaxis]

    _Yhat = Yhat * Y
    _Yhat /= _Yhat.sum(axis=1)[:, np.newaxis]
    Yhat -= _Yhat
    grad = np.dot(X.T, Yhat)
    grad /= -float(N)
    grad = grad.ravel()

    return grad



"""
Warping Invariant PCR Regression using SRSF

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

import numpy as np
import fdasrsf as fs
import fdasrsf.utility_functions as uf
import fdasrsf.curve_functions as cf
import fdasrsf.regression as rg
import fdasrsf.geometry as geo
from scipy.linalg import inv


class elastic_curve_pcr_regression:
    """
    This class provides elastic curve pcr regression for functional data using the
    SRVF framework accounting for warping

    Usage:  obj = elastic_curve_pcr_regression(beta,y)

    :param beta: numpy ndarray of shape (n, M, N) describing N curves in R^M
    :param y: response vector of length N
    :param warp_data: fdacurve object of alignment and shape PCA
    :param alpha: intercept
    :param b: coefficient vector
    :param SSE: sum of squared errors

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  29-Oct-2021
    """

    def __init__(self, beta, y):
        """
        Construct an instance of the elastic_curve_pcr_regression class
        :param beta: numpy ndarray of shape (n, M, N) describing N curves in R^M
        :param y: response vector
        :param time: vector of size M describing the sample points
        """

        self.beta = beta
        self.y = y

    def calc_model(self, no=5, T=None, rotation=True, parallel=False):
        """
        This function identifies a regression model with phase-variability
        using elastic pca

        :param no: scalar specify number of principal components (default=5)
        :param T: number of resample curve to
        :param rotation: include rotation (default = T)
        :param parallel: run in parallel (default = F)
        """

        self.rotation = rotation
        M = self.beta.shape[1]
        if T is None:
            T = M

        # Align Data
        self.warp_data = fs.fdacurve(self.beta, N=T)
        self.warp_data.karcher_mean(rotation=rotation, parallel=parallel)

        # Calculate PCA
        self.warp_data.shape_pca(no=no)

        # OLS using PCA basis
        lam = 0
        R = 0
        N1 = self.warp_data.coef.shape[1]
        Phi = np.ones((N1, no + 1))
        Phi[:, 1 : (no + 1)] = self.warp_data.coef.T
        xx = np.dot(Phi.T, Phi)
        inv_xx = inv(xx + lam * R)
        xy = np.dot(Phi.T, self.y)
        b = np.dot(inv_xx, xy)
        alpha = b[0]
        b = b[1 : no + 1]

        # compute the SSE
        int_X = np.zeros(N1)
        for ii in range(0, N1):
            int_X[ii] = np.sum(self.warp_data.coef[:, ii] * b)

        SSE = np.sum((self.y - alpha - int_X) ** 2)

        self.alpha = alpha
        self.b = b
        self.SSE = SSE

        return

    def predict(self, newdata=None):
        """
        This function performs prediction on regression model on new data if available or current stored data in object
        Usage:  obj.predict()
                obj.predict(newdata)

        :param newdata: dict containing new data for prediction (needs the keys below, if None predicts on training data)
        :type newdata: dict
        :param beta: (n, M,N) matrix of curves
        :param y: truth if available
        """

        T = self.warp_data.beta_mean.shape[1]
        if newdata != None:
            beta = newdata["beta"]
            y = newdata["y"]
            n = beta.shape[2]
            beta1 = np.zeros(beta.shape)
            q = np.zeros(beta.shape)
            for ii in range(0, n):
                if beta.shape[1] != T:
                    beta1[:, :, ii] = cf.resamplecurve(beta[:, :, ii], T)
                else:
                    beta1[:, :, ii] = beta[:, :, ii]
                a = -cf.calculatecentroid(beta1[:, :, ii])
                beta1[:, :, ii] += np.tile(a, (T, 1)).T
                q[:, :, ii] = cf.curve_to_q(beta1[:, :, ii])[0]

            mu = self.warp_data.q_mean

            v = np.zeros(q.shape)
            for ii in range(0, n):
                qn_t, R, gamI = cf.find_rotation_and_seed_unique(
                    mu, q[:, :, ii], 0, self.rotation
                )
                qn_t = qn_t / np.sqrt(cf.innerprod_q2(qn_t, qn_t))

                q1dotq2 = cf.innerprod_q2(mu, qn_t)

                if q1dotq2 > 1:
                    q1dotq2 = 1

                d = np.arccos(q1dotq2)

                u = qn_t - q1dotq2 * mu
                normu = np.sqrt(cf.innerprod_q2(u, u))
                if normu > 1e-4:
                    v[:, :, ii] = u * np.arccos(q1dotq2) / normu
                else:
                    v[:, :, ii] = np.zeros(qn_t.shape)

            Utmp = self.warp_data.U.T
            no = self.warp_data.U.shape[1]
            VM = np.mean(self.warp_data.v, 2)
            VM = VM.flatten()

            x = np.zeros((no, n))
            for i in range(0, n):
                tmp = v[:, :, i]
                tmpv1 = tmp.flatten()
                x[:, i] = Utmp.dot((tmpv1 - VM))

            self.y_pred = np.zeros(n)
            for ii in range(0, n):
                self.y_pred[ii] = self.alpha + np.dot(x[:, ii], self.b)

            if y is None:
                self.SSE = np.nan
            else:
                self.SSE = np.sum((y - self.y_pred) ** 2)
        else:
            n = self.warp_data.coef.shape[1]
            self.y_pred = np.zeros(n)
            for ii in range(0, n):
                self.y_pred[ii] = self.alpha + np.dot(
                    self.warp_data.coef[:, ii], self.b
                )

            self.SSE = np.sum((self.y - self.y_pred) ** 2)

        return



"""
geometry functions for SRSF Manipulations

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

from numpy import arccos, sin, cos, linspace, zeros, sqrt, newaxis
from numpy import ones, diff, gradient, log, exp, logspace, any
from scipy.interpolate import UnivariateSpline
from scipy.integrate import trapezoid, cumulative_trapezoid
import fdasrsf.utility_functions as uf

def inv_exp_map(Psi, psi):
    tmp = inner_product(Psi, psi)
    if tmp > 1:
        tmp = 1
    if tmp < -1:
        tmp = -1

    theta = arccos(tmp)

    if theta < 1e-10:
        exp_inv = zeros(psi.shape[0])
    else:
        exp_inv = theta / sin(theta) * (psi - cos(theta) * Psi)

    return exp_inv, theta


def exp_map(psi, v):
    v_norm = L2norm(v)
    if v_norm.sum() == 0:
        expgam = cos(v_norm) * psi
    else:
        expgam = cos(v_norm) * psi + sin(v_norm) * v / v_norm
    return expgam


def inner_product(psi1, psi2):
    M = psi1.shape[0]
    t = linspace(0, 1, M)
    ip = trapezoid(psi1 * psi2, t)
    return ip


def L2norm(psi):
    M = psi.shape[0]
    t = linspace(0, 1, M)
    l2norm = sqrt(trapezoid(psi * psi, t))
    return l2norm


def gam_to_h(gam, smooth=True):
    TT = gam.shape[0]
    time = linspace(0, 1, TT)
    binsize = diff(time)
    binsize = binsize.mean()
    if gam.ndim == 1:
        if smooth:
            gamtmp = uf.smooth_data(gam[:, newaxis], 25)
            psi = log(gradient(gamtmp[:, 0], binsize))
            h = psi - trapezoid(psi, time)
        else:
            psi = log(gradient(gam, binsize))
            h = psi - trapezoid(psi, time)
    else:
        n = gam.shape[1]
        if smooth:
            gamtmp = uf.smooth_data(gam, 25)

        psi = zeros((TT, n))
        for i in range(0, n):
            if smooth:
                psi[:, i] = log(gradient(gamtmp[:, i], binsize))
            else:
                psi[:, i] = log(gradient(gam[:, i], binsize))

        h = zeros((TT, n))
        for i in range(0, n):
            h[:, i] = psi[:, i] - trapezoid(psi[:, i], time)

    return h


def gam_to_v(gam, smooth=True):
    TT = gam.shape[0]
    time = linspace(0, 1, TT)
    binsize = diff(time)
    binsize = binsize.mean()
    mu = ones(TT)
    if gam.ndim == 1:
        if smooth:
            tmp_spline = UnivariateSpline(time, gam, s=1e-4)
            g = tmp_spline(time, 1)
            idx = g <= 0
            g[idx] = 0
            psi = sqrt(g)
        else:
            psi = sqrt(gradient(gam, binsize))
        vec, theta = inv_exp_map(mu, psi)
    else:
        n = gam.shape[1]

        psi = zeros((TT, n))
        for i in range(0, n):
            if smooth:
                tmp_spline = UnivariateSpline(time, gam[:, i], s=1e-4)
                g = tmp_spline(time, 1)
                idx = g <= 0
                g[idx] = 0
                psi[:, i] = sqrt(g)
            else:
                psi[:, i] = sqrt(gradient(gam[:, i], binsize))

        vec = zeros((TT, n))
        for i in range(0, n):
            out, theta = inv_exp_map(mu, psi[:, i])
            vec[:, i] = out

    return vec


def h_to_gam(h):
    TT = h.shape[0]
    time = linspace(0, 1, TT)
    if h.ndim == 1:
        gam0 = cumulative_trapezoid(exp(h), time, initial=0)
        gam0 /= trapezoid(exp(h), time)
        gam = (gam0 - gam0.min()) / (gam0.max() - gam0.min())
    else:
        n = h.shape[1]

        gam = zeros((TT, n))
        for i in range(0, n):
            gam0 = cumulative_trapezoid(exp(h[:, i]), time, initial=0)
            gam0 /= trapezoid(exp(h[:, i]), time)
            gam[:, i] = (gam0 - gam0.min()) / (gam0.max() - gam0.min())

    return gam


def v_to_gam(v):
    TT = v.shape[0]
    time = linspace(0, 1, TT)
    mu = ones(TT)
    if v.ndim == 1:
        psi = exp_map(mu, v)
        gam0 = cumulative_trapezoid(psi * psi, time, initial=0)
        gam = (gam0 - gam0.min()) / (gam0.max() - gam0.min())
    else:
        n = v.shape[1]

        gam = zeros((TT, n))
        for i in range(0, n):
            psi = exp_map(mu, v[:, i])
            gam0 = cumulative_trapezoid(psi * psi, time, initial=0)
            gam[:, i] = (gam0 - gam0.min()) / (gam0.max() - gam0.min())

    return gam



import unittest
import numpy as np
import fdasrsf as fs


class TestFDASRSF(unittest.TestCase):
    # Returns True or False.
    def test_reparm(self):
        M = 101
        q1 = np.sin(np.linspace(0, 2 * np.pi, M))
        timet = np.linspace(0, 1, M)
        gam = fs.optimum_reparam(q1, timet, q1)
        self.assertAlmostEqual(sum(gam - timet), 0)

    def test_warp_f_gamma(self):
        M = 101
        q1 = np.sin(np.linspace(0, 2 * np.pi, M))
        timet = np.linspace(0, 1, M)
        gam = fs.optimum_reparam(q1, timet, q1)
        q1a = fs.warp_f_gamma(timet, q1, gam)
        self.assertAlmostEqual(sum(q1 - q1a), 0)

    def test_warp_q_gamma(self):
        M = 101
        q1 = np.sin(np.linspace(0, 2 * np.pi, M))
        timet = np.linspace(0, 1, M)
        gam = fs.optimum_reparam(q1, timet, q1)
        q1a = fs.warp_q_gamma(timet, q1, gam)
        self.assertAlmostEqual(sum(q1 - q1a), 0)

    def test_reparm1(self):
        M = 101
        q1 = np.sin(np.linspace(0, 2 * np.pi, M))
        timet = np.linspace(0, 1, M)
        gam = fs.optimum_reparam(q1, timet, q1, method="DP")
        self.assertAlmostEqual(sum(gam - timet), 0)

    def test_rlbgs(self):
        M = 101
        q1 = np.sin(np.linspace(0, 2 * np.pi, M))
        timet = np.linspace(0, 1, M)
        gam = fs.optimum_reparam(q1, timet, q1, method="RBFGS")
        self.assertAlmostEqual(sum(gam - timet), 0)

    def test_crlbgs(self):
        M = 101
        q1 = np.sin(np.linspace(0, 2 * np.pi, M))
        timet = np.linspace(0, 1, M)
        gam = fs.optimum_reparam(q1, timet, q1, method="cRBFGS")
        self.assertAlmostEqual(sum(gam - timet), 0)

    def test_f_to_srvf(self):
        M = 101
        f1 = np.sin(np.linspace(0, 2 * np.pi, M))
        timet = np.linspace(0, 1, M)
        q1 = fs.f_to_srsf(f1, timet)
        f1a = fs.srsf_to_f(q1, timet)
        self.assertAlmostEqual(sum(f1 - f1a), 0, 4)

    def test_elastic_distance(self):
        M = 101
        f1 = np.sin(np.linspace(0, 2 * np.pi, M))
        timet = np.linspace(0, 1, M)
        da, dp = fs.elastic_distance(f1, f1, timet)
        self.assertLessEqual(da, 1e-10)
        self.assertLessEqual(dp, 1e-6)

    def test_smooth(self):
        M = 101
        q1 = np.zeros((M, 1))
        q1[:, 0] = np.sin(np.linspace(0, 2 * np.pi, M)).T
        q1a = fs.smooth_data(q1, 1)
        q1b = fs.smooth_data(q1, 1)
        self.assertAlmostEqual(sum(q1a.flatten() - q1b.flatten()), 0)

    def test_edistance(self):
        M = 101
        q1 = np.sin(np.linspace(0, 2 * np.pi, M))
        timet = np.linspace(0, 1, M)
        dy, dx = fs.elastic_distance(q1, q1, timet)
        self.assertAlmostEqual(dy + dx, 0)

    def test_invgamma(self):
        M = 101
        gam = np.linspace(0, 1, M)
        gami = fs.invertGamma(gam)
        self.assertAlmostEqual(sum(gam - gami), 0)

    def test_invexpmap(self):
        M = 101
        gam = np.linspace(0, 1, M)
        binsize = np.mean(np.diff(gam))
        psi = np.sqrt(np.gradient(gam, binsize))
        out, theta = fs.inv_exp_map(psi, psi)
        self.assertAlmostEqual(sum(out), 0)

    def test_l2norm(self):
        M = 101
        gam = np.linspace(0, 1, M)
        binsize = np.mean(np.diff(gam))
        psi = np.sqrt(np.gradient(gam, binsize))
        out, theta = fs.inv_exp_map(psi, psi)
        out1 = fs.geometry.L2norm(out)
        self.assertAlmostEqual(out1, 0)

    def test_expmap(self):
        M = 101
        gam = np.linspace(0, 1, M)
        binsize = np.mean(np.diff(gam))
        psi = np.sqrt(np.gradient(gam, binsize))
        out, theta = fs.inv_exp_map(psi, psi)
        out1 = fs.exp_map(psi, out)
        self.assertAlmostEqual(sum(out1), M)

    def test_srsf(self):
        data = np.load("bin/simu_data.npz")
        time = data["arr_1"]
        f = data["arr_0"]
        obj = fs.fdawarp(f, time)
        obj.srsf_align()
        vpca = fs.fdavpca(obj)
        vpca.calc_fpca()
        vpca = fs.fdajpca(obj)
        vpca.calc_fpca()
        vpca = fs.fdahpca(obj)
        vpca.calc_fpca()
        self.assertAlmostEqual(obj.amp_var, 0.018998691036349585)

    def test_srvf(self):
        data = np.load("bin/MPEG7.npz", allow_pickle=True)
        Xdata = data["Xdata"]
        curve = Xdata[0, 1]
        n, M = curve.shape
        K = Xdata.shape[1]

        beta = np.zeros((n, M, K))
        for i in range(0, K):
            beta[:, :, i] = Xdata[0, i]

        obj = fs.fdacurve(beta, N=M)
        obj.karcher_mean()
        obj.srvf_align()
        obj.karcher_cov()
        obj.shape_pca()
        self.assertAlmostEqual(obj.E[-1], 0.022668183569717587)


if __name__ == "__main__":
    unittest.main()



"""
Utility functions for SRSF Manipulations

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""
import numpy as np
import time
from scipy.integrate import trapezoid, cumulative_trapezoid
from numpy.linalg import norm


class rlbfgs:
    r"""
    This class provides alignment methods for functional data using the SRVF
    framework using the Riemannian limited memory BFGS solver.  The solver is
    designed to operate on the positive orthant of the unit hypersphere in
    :math:`L^2([0,1],R)`. The set of all functions
    :math:`h=\sqrt{\dot{\gamma}}`,  where :math:`\gamma` is a diffeomorphism,
    is that manifold.

    The inputs q1 and q2 are the square root velocity functions of curves in
    R^n to be aligned. Here, q2 will be aligned to q1.

    Usage:  obj = rlbfgs(q1,q2,t)

    :param q1: (M,N): matrix defining srvf of dimension M of N samples
    :param q2: (M,N): matrix defining srvf of dimension M of N samples
    :param t: time vector of length N
    :param q2Opt: optimally aligned srvf
    :param gammaOpt: optimal warping function
    :param cost: final cost
    :param info: dictionary consisting of info about the iterations

    Wen Huang, Kyle A. Gallivan, Anuj Srivastava, Pierre-Antoine Absil.
    "Riemannian Optimization for Elastic Shape Analysis", Short version,
    The 21st International Symposium on Mathematical Theory of Networks and
    Systems (MTNS 2014).

    Code based on rlbfgs.m in Manopt: www.manopt.org

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  27-Oct-2020
    """

    def __init__(self, q1, q2, t):
        """
        Construct an instance of the rlbfgs class
        :param q1: (M,N): matrix defining srvf of dimension M of N samples
        :param q2: (M,N): matrix defining srvf of dimension M of N samples
        :param t: time vector of length N
        """

        self.t = t
        self.T = t.shape[0]
        if q1.ndim > 1:
            self.q1 = q1
            self.q2 = q2
        else:
            self.q1 = q1 / norm(q1)
            self.q2 = q2 / norm(q2)

    def solve(self, maxiter=30, verb=0, lam=0, penalty="roughness"):
        """
        Run solver

        :param maxiter: maximum number of interations
        :param verb: integer used to tune the amount of output
        :param lam: amount of penalty
        :param penalty: penalty, "roughness", "l2gam", "l2psi", "geodesic"
        """

        # @todo add options to parameters if needed
        # terminates if the norm of the gradient drops below this
        tolgradnorm = 1e-3
        # terminates if more than seconds elapsed
        maxtime = np.inf
        # minimum norm of tangent vector that points from current to next
        minstepsize = 1e-10
        # number of previous iterations the program remembers
        memory = 30
        memory = max(memory, 0)
        # the cautious step needs a real function that has value 0 at t=0
        strict_inc_func = lambda t: 1e-4 * t
        ls_max_steps = 25

        options = {
            "tolgradnorm": tolgradnorm,
            "maxtime": maxtime,
            "memory": memory,
            "strict_inc_func": strict_inc_func,
            "ls_max_steps": ls_max_steps,
            "maxiter": maxiter,
            "minstepsize": minstepsize,
        }

        timetic = time.time()
        # Initialization of Variables
        htilde = np.ones(self.T)
        q2tilde = self.q2

        # number of iterations since last restart
        j = 0

        # Total number of BFGS iterations
        k = 0

        # list to store step vectors which point from h_id to h_{k+1}
        # for k indexing the last iterations, capped at option memory
        sHistory = [None] * memory

        # list store differences for latest k's for the gradient at time
        # k+1 and the gradient at time k
        yHistory = [None] * memory

        # stores the reciprocal of the inner product between
        # sHistory[k] and yHistory[k]
        rhoHistory = [None] * memory

        # scaling of direction given by getDirection
        alpha = 1

        # scaling of initial matrix, Barzilai-Borwein
        scaleFactor = 1

        # Norm of the step
        stepsize = 1

        # sores wether the step is accepted byt he cautious update check
        accepted = True

        # compute cost function and its gradient
        hCurCost, hCurGradient = self.alignment_costgrad(q2tilde, htilde, lam, penalty)
        hCurGradNorm = self.norm(hCurGradient)

        # line-search statistics for recording in info
        lsstats = {"costevals": 0, "stepsize": 0.0, "alpha": 0.0}

        # flag to control restarting scheme to avoid infinite loops
        ultimatum = False

        if verb >= 2:
            print(
                " iter                   cost val            grad. norm           alpha\n"
            )

        # stats
        info = []
        stats = {
            "iter": k,
            "cost": hCurCost,
            "gradnorm": hCurGradNorm,
            "stepsize": np.nan,
            "time": time.time() - timetic,
            "accepted": None,
            "linesearch": lsstats,
        }
        info.append(stats)

        while True:
            if verb >= 2:
                print(
                    "%5d    %+.16e        %.8e      %.4e\n"
                    % (k, hCurCost, hCurGradNorm, alpha)
                )

            # Start timing this iteration
            timetic = time.time()

            # run standard stopping criterion checks
            stop = self.stoppingcriterion(options, info, k)

            if stop == 0:
                if stats["stepsize"] < options["minstepsize"]:
                    if not ultimatum:
                        if verb >= 2:
                            print(
                                "stepsize is too small, restarting the bfgs procedure at the current point.\n"
                            )
                        j = 0
                        ultimatum = True
                    else:
                        stop = 1
                else:
                    # we are not in trouble: list the ultimatum if it was on
                    ultimatum = False

            if stop > 0:
                break

            # compute BFGS direction
            p = self.getDirection(
                hCurGradient,
                sHistory,
                yHistory,
                rhoHistory,
                scaleFactor,
                min(j, memory),
            )

            # execute line-search
            in_prod = self.inner(hCurGradient, p)
            stepsize, hNext, lsstats = self.linesearch_hint(
                p, hCurCost, in_prod, q2tilde, lam, penalty
            )

            # Iterative update of optimal diffeomorphism and q2 via group
            # action
            htilde = self.group_action_SRVF(htilde, hNext)
            q2tilde = self.group_action_SRVF(q2tilde, hNext)

            # Record the BFGS step-multiplier alpha which was effectively
            # selected. Toward convergence, we hope to see alpha = 1.
            alpha = stepsize / self.norm(p)
            step = alpha * p

            # query cost and gradient at the candidate new point
            hNextCost, hNextGradient = self.alignment_costgrad(
                q2tilde, hNext, lam, penalty
            )

            # compute sk and yk
            sk = step
            yk = hNextGradient - hCurGradient

            # Computation of the BFGS step is invariant under scaling of sk and
            # yk by a common factor. For numerical reasons, we scale sk and yk
            # so that sk is a unit norm vector.
            norm_sk = self.norm(sk)
            sk = sk / norm_sk
            yk = yk / norm_sk

            inner_sk_yk = self.inner(sk, yk)
            inner_sk_sk = self.norm(sk) ** 2  # ensures nonnegativity

            # If the cautious step is accepted (which is the intended
            # behavior), we record sk, yk, and rhok and need to do some
            # housekeeping. If the cautious step is rejected, these are not
            # recorded. In all cases, hNext is the next iterate: the notion of
            # accept/reject here is limited to whether or not we keep track of
            # sk, yk, rhok to update the BFGS operator.
            cap = options["strict_inc_func"](hCurGradNorm)
            if inner_sk_sk != 0 and (inner_sk_yk / inner_sk_sk) >= cap:
                accepted = True

                rhok = 1 / inner_sk_yk

                scaleFactor = inner_sk_yk / self.norm(yk) ** 2

                # Time to store the vectors sk, yk and the scalar rhok

                # If we are out of memory
                if j >= memory:
                    # sk and yk are saved from 1 to the end with the most
                    # current recorded to the rightmost hand side of the cells
                    # that are occupied. When memory is full, do a shift so
                    # that the rightmost is earliest and replace it with the
                    # most recent sk, yk.
                    if memory > 1:
                        tmp = sHistory[1:]
                        tmp.append(sHistory[0])
                        sHistory = tmp
                        tmp = yHistory[1:]
                        tmp.append(yHistory[0])
                        yHistory = tmp
                        tmp = rhoHistory[1:]
                        tmp.append(rhoHistory[0])
                        rhoHistory = tmp
                    if memory > 0:
                        sHistory[memory] = sk
                        yHistory[memory] = yk
                        rhoHistory[memory] = rhok
                    # if we are not out of memory
                else:
                    sHistory[j] = sk
                    yHistory[j] = yk
                    rhoHistory[j] = rhok

                j += 1

                # the cautious step is rejected we do not store sk, yk, and r
                # hok
            else:
                accepted = False

            # update variables to new iterate
            hCurGradient = hNextGradient
            hCurGradNorm = self.norm(hNextGradient)
            hCurCost = hNextCost

            # iter is the number of iterations we have accomplished.
            k += 1
            stats = {
                "iter": k,
                "cost": hCurCost,
                "gradnorm": hCurGradNorm,
                "stepsize": np.nan,
                "time": time.time() - timetic,
                "accepted": accepted,
                "linesearch": lsstats,
            }
            info.append(stats)

        self.info = info[0 : (k + 1)]
        self.gammaOpt = np.zeros(self.T)
        self.gammaOpt[1:] = cumulative_trapezoid(htilde**2, self.t)
        self.q2Opt = q2tilde
        self.cost = hCurCost

        if verb >= 1:
            print("Total time is %f [s] (excludes statsfun)\n" % info[-1].time)

        return

    def alignment_cost(self, h, q2k, lam=0, penalty="roughness"):
        r"""
        Evaluate the cost function :math:`f = ||q1 - ((q2,hk),h)||^2`.
        :math:`h=sqrt{\dot{\gamma}}` is a sequential update of cumulative
        warping hk
        """
        q2new = self.group_action_SRVF(q2k, h)

        if penalty == "roughness":
            time1 = np.linspace(0, 1, h.shape[0])
            binsize = np.mean(np.diff(time1))
            pen = trapezoid(np.gradient(h**2, binsize) ** 2, time1)
        elif penalty == "l2gam":
            pen = self.normL2(h**2 - np.ones(h.shape[0])) ** 2
        elif penalty == "l2psi":
            pen = self.normL2(h - np.ones(h.shape[0])) ** 2
        elif penalty == "geodesic":
            time1 = np.linspace(0, 1, h.shape[0])
            q1dotq2 = trapezoid(h, time1)
            if q1dotq2 > 1:
                q1dotq2 = 1
            elif q1dotq2 < -1:
                q1dotq2 = -1
            pen = np.real(np.arccos(q1dotq2)) ** 2
        else:
            raise Exception("penalty not implemented")

        f = self.normL2(self.q1 - q2new) ** 2 + lam * pen

        return f

    def alignment_costgrad(self, q2k, h, lam=0, penalty="roughness"):
        r"""
        Evaluate the cost function :math:`f = ||q1 - (q2,hk)||^2`, and
        evaluate the gradient g = grad f in the tangent space of identity.
        :math:`hk=sqrt{\dot{\gamma_k}}` is the cumulative warping of q2
        produced by an iterative sequential optimization algorithm.
        """

        if penalty == "roughness":
            time1 = np.linspace(0, 1, h.shape[0])
            binsize = np.mean(np.diff(time1))
            pen = trapezoid(np.gradient(h**2, binsize) ** 2, time1)
        elif penalty == "l2gam":
            pen = self.normL2(h**2 - np.ones(h.shape[0])) ** 2
        elif penalty == "l2psi":
            pen = self.normL2(h - np.ones(h.shape[0])) ** 2
        elif penalty == "geodesic":
            time1 = np.linspace(0, 1, h.shape[0])
            q1dotq2 = trapezoid(h, time1)
            if q1dotq2 > 1:
                q1dotq2 = 1
            elif q1dotq2 < -1:
                q1dotq2 = -1
            pen = np.real(np.arccos(q1dotq2)) ** 2
        else:
            raise Exception("penalty not implemented")

        t = self.t
        T = self.T
        q1 = self.q1

        # compute cost
        f = self.normL2(q1 - q2k) ** 2 + lam * pen

        # compute cost gradient
        q2kdot = np.gradient(q2k, 1 / (T - 1))
        if q2k.ndim > 1:
            q2kdot = q2kdot[1]
        dq = q1 - q2k
        v = np.zeros(T)
        tmp = dq * q2kdot
        tmp1 = dq * q2k
        if tmp.ndim > 1:
            v[1:] = 2 * cumulative_trapezoid(tmp.sum(axis=0), t)
            v = v - tmp1.sum(axis=0)
        else:
            v[1:] = 2 * cumulative_trapezoid(tmp, t)
            v = v - tmp1

        g = v - trapezoid(v, t)

        return f, g

    def getDirection(
        self, hCurGradient, sHistory, yHistory, rhoHistory, scaleFactor, j
    ):
        """
        BFGS step, see Wen's paper for details. This function takes in a
        tangent vector g, and applies an approximate inverse Hessian P to
        it to get Pg. Then, -Pg is returned. Parallel transport is not
        needed for this problem since we always work in the tangent space
          of identity.
        """
        q = hCurGradient
        inner_s_q = np.zeros(j)

        for i in range(j, 0, -1):
            inner_s_q[i - 1] = rhoHistory[i - 1] * self.inner(sHistory[i - 1], q)
            q = q - inner_s_q[i - 1] * yHistory[i - 1]

        r = scaleFactor * q

        for i in range(0, j):
            omega = rhoHistory[i] * self.inner(yHistory[i], r)
            r = r + (inner_s_q[i] - omega) * sHistory[i]

        direction = -r

        return direction

    def linesearch_hint(self, d, f0, df0, q2k, lam=0, penalty="roughness"):
        """
        Armijo line-search based on the line-search hint in the problem
        structure.

        Base line-search algorithm for descent methods, based on a simple
        backtracking method. The search direction provided has to be a descent
        direction, as indicated by a negative df0 = directional derivative of f
        at the identity element along d.

        The algorithm selects a hardcoded initial step size. If that
        step does not fulfill the Armijo sufficient decrease criterion, that
        step size is reduced geometrically until a satisfactory step size is
        obtained or until a failure criterion triggers.

        Below, the step is constructed as alpha*d, and the step size is the
        norm of that vector, thus: stepsize = alpha*norm_d. The step is
        executed by computing the exponential mapping exp_{hid}(alpha*d),
        giving newh.
        """
        contraction_factor = 0.5
        suff_decr = 1e-6
        max_ls_steps = 25
        ls_backtrack = True
        ls_force_decrease = True

        # init alpha
        alpha = 1

        # Identity element
        hid = np.ones(self.T)

        # Make the chosen step and compute cost there
        newh = self.exp(hid, d, alpha)
        newf = self.alignment_cost(newh, q2k, lam, penalty)
        cost_evaluations = 1

        # backtrack while the Armijo criterion is not satisfied
        # or if newh goes outside positive orthant
        tst = newh <= 0
        while ls_backtrack and (
            (newf > (f0 + suff_decr * alpha * df0)) or (tst.sum() > 0)
        ):
            # reduce the step size
            alpha *= contraction_factor

            # look closer down the line
            newh = self.exp(hid, d, alpha)
            newf = self.alignment_cost(newh, q2k, lam, penalty)
            cost_evaluations += 1
            tst = newh <= 0

            # make sure we don't run out of budget
            if cost_evaluations >= max_ls_steps:
                break

        # if we got here with obtaining a derease, reject the step
        if ls_force_decrease and newf > f0:
            alpha = 0
            newh = hid
            newf = f0

        # As seen outside this function, stepsize is the size of the vector we
        # retract to make the step from h to newh. Since the step is alpha*d:
        norm_d = self.norm(d)
        stepsize = alpha * norm_d

        # return some statistics
        lsstats = {"costevals": cost_evaluations, "stepsize": stepsize, "alpha": alpha}

        return stepsize, newh, lsstats

    def stoppingcriterion(self, options, info, last):
        stop = 0
        stats = info[last]
        if stats["gradnorm"] <= options["tolgradnorm"]:
            stop = 2

        if stats["time"] >= options["maxtime"]:
            stop = 3

        if stats["iter"] >= options["maxiter"]:
            stop = 4

        return stop

    def group_action_SRVF(self, q, h):
        p = q.shape[0]
        gamma = np.zeros(self.T)
        gamma[1:] = cumulative_trapezoid(h**2, self.t)
        gamma = gamma / gamma[-1]
        h = np.sqrt(np.gradient(gamma, self.t))
        qnew = np.zeros(q.shape)
        if q.ndim > 1:
            for i in range(0, p):
                qnew[i, :] = np.interp(gamma, self.t, q[i, :]) * h
        else:
            qnew = np.interp(gamma, self.t, q) * h

        return qnew

    def normL2(self, f):
        val = np.sqrt(self.innerProdL2(f, f))
        return val

    def innerProdL2(self, f1, f2):
        tmp = f1 * f2
        if tmp.ndim > 1:
            val = trapezoid(tmp.sum(axis=0), self.t)
        else:
            val = trapezoid(tmp, self.t)
        return val

    def dist(self, f1, f2):
        d = np.real(np.arccos(self.inner(f1, f2)))
        return d

    def typicaldist(self):
        return np.pi / 2

    def proj(self, f, v):
        out = v - f * trapezoid(f * v, self.t)
        return out

    def log(self, f1, f2):
        v = self.proj(f1, f2 - f1)
        di = self.dist(f1, f2)
        if di > 1e-6:
            nv = self.norm(v)
            v = v * (di / nv)

        return v

    def exp(self, f1, v, delta=1):
        vd = delta * v
        nrm_vd = self.norm(vd)

        # Former versions of Manopt avoided the computation of sin(a)/a for
        # small a, but further investigations suggest this computation is
        # well-behaved numerically.
        if nrm_vd > 0:
            f2 = f1 * np.cos(nrm_vd) + vd * (np.sin(nrm_vd) / nrm_vd)
        else:
            f2 = f1

        return f2

    def transp(self, f1, f2, v):
        """
        Isometric vector transport of d from the tangent space at x1 to x2.
        This is actually a parallel vector transport, see (5) in
        http://epubs.siam.org/doi/pdf/10.1137/16M1069298
        "A Riemannian Gradient Sampling Algorithm for Nonsmooth Optimization
        on Manifolds", by Hosseini and Uschmajew, SIOPT 2017
        """
        w = self.log(f1, f2)
        dist_f1f2 = self.norm(w)
        if dist_f1f2 > 0:
            u = w / dist_f1f2
            utv = self.inner(u, v)
            Tv = v + (np.cos(dist_f1f2) - 1) * utv * u - np.sin(dist_f1f2) * utv * f1
        else:
            Tv = v

        return Tv

    def inner(self, v1, v2):
        return trapezoid(v1 * v2, self.t)

    def norm(self, v):
        return np.sqrt(trapezoid(v**2, self.t))

    def zerovec(self):
        return np.zeros(self.T)



"""
Elastic functional change point detection

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

import numpy as np
import matplotlib.pyplot as plt
import fdasrsf as fs
import fdasrsf.fPCA as fpca
import fdasrsf.utility_functions as uf
from scipy.linalg import norm, svd
from fdasrsf.geometry import L2norm


# Function Definitions
def BBridge(x=0, y=0, t0=0, T=1, N=100):
    if T <= t0:
        raise Exception("Endpoint is earlier than beginning")

    dt = (T - t0) / N
    t = np.linspace(t0, T, N + 1)
    rng = np.random.default_rng()
    samp = rng.standard_normal(N)
    X = np.insert(np.cumsum(samp) * np.sqrt(dt), 0, 0)
    BB = x + X - (t - t0) / (T - t0) * (X[N] - y + x)
    return BB


def LongRunCovMatrixPrecentered(mdobj, h=0, kern_type="bartlett"):
    D = mdobj.shape[0]
    N = mdobj.shape[1]

    def Kernel(i, h):
        x = i / h
        if kern_type == "flat":
            return 1
        if kern_type == "simple":
            return 0
        if kern_type == "bartlett":
            return 1 - x
        if kern_type == "flat_top":
            if x < 0.1:
                return 1
            else:
                if x >= 0.1 & x < 1.1:
                    return 1.1 - x
                else:
                    return 0
        if kern_type == "parzen":
            if x < 0.5:
                return 1 - 6 * pow(x, 2) + 6 * pow(abs(x), 3)
            else:
                return 2 * pow((1 - abs(x)), 3)

    D_mat = np.zeros((D, D))
    cdata = mdobj

    for k in range(0, D):
        for r in range(k, D):
            s = cdata[k, :] @ cdata[r, :]
            if h > 0:
                for i in range(h):
                    a = cdata[k, range(N - i)] @ cdata[r, range(i, N)]
                    a = a + cdata[r, range(N - i)] @ cdata[k, range(i, N)]
                    s = s + Kernel(i + 1, h) * a
            D_mat[k, r] = s
            D_mat[r, k] = D_mat[k, r]
    return D_mat / N


class elastic_change:
    """ "
    This class provides elastic changepoint using elastic fpca

    Usage:  obj = elastic_change(f,time)

    :param f: (M,N) % matrix defining N functions of M samples
    :param time: time vector of length M
    :param BBridges: precomputed Brownian Bridges (default: None)
    :param use_BBridges: use precomputed Brownian Bridges (default: False)
    :param warp_data: aligned data (default: None)
    :param Sn: test statistic values
    :param Tn: max of test statistic
    :param p: p-value
    :param k_star: change point
    :param values: values of computed Brownian Bridges
    :param dat_a: data before changepoint
    :param dat_b: data after changepoint
    :param warp_a: warping functions before changepoint
    :param warp_b: warping functions after changepoint
    :param mean_a: mean function before changepoint
    :param mean_b: mean function after changepoint
    :param warp_mean_a: mean warping function before changepoint
    :param warp_mean_b: mean warping function after changepoin

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  27-Apr-2022

    """

    def __init__(
        self,
        f,
        time,
        BBridges=None,
        use_BBridges=False,
        smooth_data=False,
        warp_data=None,
        use_warp_data=False,
        parallel=False,
        sparam=25,
    ):
        """
        Construct an instance of the elastic_change class
        :param f: (M,N) % matrix defining N functions of M samples
        :param time: time vector of length M
        :param BBridges: precomputed Brownian Bridges (default: None)
        :param use_BBridges: use precomputed Brownian Bridges (default: False)
        :param smooth_data: smooth function data (default: False)
        :param warp_data: precomputed aligned data (default: None)
        :param use_warp_data: use precomputed warping data (default: False)
        :param parallel: run computation in parallel (default: True)
        :param sparam: number of smoothing runs of box filter (default: 25)
        """
        self.f = f
        self.time = time
        self.BBridges = BBridges
        self.use_BBridges = use_BBridges

        if smooth_data:
            self.f = fs.smooth_data(self.f, sparam)

        if use_warp_data:
            self.warp_data = warp_data  # Align Data
        else:
            self.warp_data = fs.fdawarp(self.f, self.time)
            self.warp_data.srsf_align(parallel=parallel)

    def compute(
        self,
        pca_method="vert",
        pc=0.95,
        d=1000,
        compute_epidemic=False,
        n_pcs=5,
        preset_pcs=False,
    ):
        """
        Compute elastic change detection

        :param pca_method: string specifying pca method (options = "combined","vert", or "horiz", default = "combined")
        :param pc: percentage of cumulative variance to use (default: 0.95)
        :param compute_epidemic: compute epidemic changepoint model (default: False)
        :param n_pcs: scalar specify number of principal components (default: 5)
        :param preset_pcs: use all PCs (default: False)

        """
        N1 = self.f.shape[1]

        # Calculate PCA
        if pca_method == "combined":
            self.pca = fpca.fdajpca(self.warp_data)
        elif pca_method == "vert":
            self.pca = fpca.fdavpca(self.warp_data)
        elif pca_method == "horiz":
            self.pca = fpca.fdahpca(self.warp_data)
        else:
            raise Exception("Invalid fPCA Method")
        self.pca.calc_fpca(N1)
        cumm_coef = np.cumsum(self.pca.latent) / sum(self.pca.latent)
        if preset_pcs == True:
            self.no = n_pcs
        else:
            self.no = np.argwhere(cumm_coef >= pc)[0][0]

        lam = 1 / self.pca.latent[0 : self.no]
        Sigma = np.diag(lam)
        eta = self.pca.coef[:, 0 : self.no]
        eta_bar = eta.sum(axis=0) / N1

        # Compute Test Statistic
        self.Sn = np.zeros(N1)
        for i in range(1, N1):
            tmp_eta = eta[0:i, :]
            tmp = tmp_eta.sum(axis=0) - tmp_eta.shape[0] * eta_bar

            self.Sn[i] = 1 / (N1) * tmp.T @ Sigma @ tmp

        self.Tn = self.Sn.mean()
        self.k_star = self.Sn.argmax()

        # compute distribution
        if self.use_BBridges == False:
            values = np.zeros(d)
            for i in range(d):
                B_tmp = np.zeros((self.no, N1))
                for j in range(self.no):
                    B_tmp[j, :] = BBridge(N=N1 - 1) ** 2
                values[i] = B_tmp.sum(axis=0).mean()
        elif self.no > self.BBridges.shape[0]:
            values = np.zeros(d)
            for i in range(d):
                B_tmp = np.zeros((self.no, N1))
                for j in range(self.no):
                    B_tmp[j, :] = BBridge(N=N1 - 1) ** 2
                values[i] = B_tmp.sum(axis=0).mean()
        else:
            values = (self.BBridges[0 : self.no, :, 0:d] ** 2).sum(axis=0).mean(axis=0)

        z = self.Tn <= values
        self.p = z[z == True].shape[0] / z.shape[0]

        self.sim_values = values

        self.dat_a = self.f[:, 0 : self.k_star]
        self.dat_b = self.f[:, self.k_star :]
        self.warp_a = self.warp_data.gam[:, 0 : self.k_star]
        self.warp_b = self.warp_data.gam[:, self.k_star :]
        self.mean_a = self.warp_data.fn[:, 0 : self.k_star].mean(axis=1)
        self.mean_b = self.warp_data.fn[:, self.k_star :].mean(axis=1)
        if self.warp_a.size == 0:
            self.warp_mean_a = np.array([])
        else:
            mu, gam_mu, psi, vec = uf.SqrtMean(self.warp_a)
            self.warp_mean_a = gam_mu

        if self.warp_b.size == 0:
            self.warp_mean_b = np.array([])
        else:
            mu, gam_mu, psi, vec = uf.SqrtMean(self.warp_b)
            self.warp_mean_b = gam_mu

        self.delta = self.mean_b - self.mean_a
        if self.warp_mean_a.size > 0 and self.warp_mean_b.size > 0:
            self.delta_warp = self.warp_mean_b - self.warp_mean_a
        else:
            self.delta_warp = self.time

        if compute_epidemic:
            self.compute_epidemic = True
            snk_minus_vals = np.zeros((N1, N1, self.no))
            snk_minus_summary = np.zeros((N1, N1))
            for i in range(N1):
                for j in range(N1):
                    if i > j:
                        continue
                    tmp_eta = eta[0:i, :]
                    tmp = tmp_eta.sum(axis=0) - tmp_eta.shape[0] * eta_bar
                    tmp_eta_j = eta[0:j, :]
                    tmp_j = tmp_eta_j.sum(axis=0) - tmp_eta.shape[0] * eta_bar
                    snk_minus_vals[i, j, :] = tmp - tmp_j
                    snk_minus_summary[i, j] = sum(pow(snk_minus_vals[i, j], 2) * lam)
            self.tau1 = np.argmax(snk_minus_summary.max(axis=1))
            self.tau2 = np.argmax(snk_minus_summary[self.tau1, :])
            self.Sn_epidemic = snk_minus_summary
            self.Tn_epidemic = 1 / (pow(N1, 3)) * snk_minus_summary.sum()
            # compute distribution

            if self.use_BBridges == False:
                values = np.zeros(d)
                for i in range(d):
                    B_tmp = np.zeros((self.no, N1, N1))
                    for j in range(self.no):
                        BBridge_val = BBridge(N=N1 - 1)
                        for i1 in range(N1):
                            for j1 in range(N1):
                                if i1 > j1:
                                    continue
                                B_tmp[j, i1, j1] = (
                                    BBridge_val[i1] - BBridge_val[j1]
                                ) ** 2
                    values[i] = B_tmp.mean(axis=(1, 2)).sum()
            elif self.no > self.BBridges.shape[0]:
                values = np.zeros(d)
                for i in range(d):
                    B_tmp = np.zeros((self.no, N1, N1))
                    for j in range(self.no):
                        BBridge_val = BBridge(N=N1 - 1)
                        for i1 in range(N1):
                            for j1 in range(N1):
                                if i1 > j1:
                                    continue
                                B_tmp[j, i1, j1] = (
                                    BBridge_val[i1] - BBridge_val[j1]
                                ) ** 2
                    values[i] = B_tmp.mean(axis=(1, 2)).sum()
            else:
                values = np.zeros(d)
                for i in range(d):
                    B_tmp = np.zeros((self.no, N1, N1))
                    for j in range(self.no):
                        for i1 in range(N1):
                            for j1 in range(N1):
                                if i1 > j1:
                                    continue
                                B_tmp[j, i1, j1] = (
                                    self.BBridges[j, i1, i] - self.BBridges[j, j1, i]
                                ) ** 2
                    values[i] = B_tmp.mean(axis=(1, 2)).sum()

            self.p_epidemic = np.mean(values >= self.Tn_epidemic)
            self.sim_values_epidemic = values

            self.dat_a_epidemic = np.column_stack(
                (self.f[:, 0 : self.tau1], self.f[:, self.tau2 :])
            )
            self.dat_b_epidemic = self.f[:, self.tau1 : self.tau2]
            self.warp_a_epidemic = np.column_stack(
                (
                    self.warp_data.gam[:, 0 : self.tau1],
                    self.warp_data.gam[:, self.tau2 :],
                )
            )
            self.warp_b_epidemic = self.warp_data.gam[:, self.tau1 : self.tau2]
            self.mean_a_epidemic = np.column_stack(
                (self.warp_data.fn[:, 0 : self.tau1], self.warp_data.fn[:, self.tau2 :])
            ).mean(axis=1)
            self.mean_b_epidemic = self.warp_data.fn[:, self.tau1 : self.tau2].mean(
                axis=1
            )
            if self.warp_a_epidemic.size == 0:
                self.warp_mean_a_epidemic = np.array([])
            else:
                mu, gam_mu, psi, vec = uf.SqrtMean(self.warp_a_epidemic)
                self.warp_mean_a_epidemic = gam_mu

            if self.warp_b_epidemic.size == 0:
                self.warp_mean_b_epidemic = np.array([])
            else:
                mu, gam_mu, psi, vec = uf.SqrtMean(self.warp_b_epidemic)
                self.warp_mean_b_epidemic = gam_mu

            self.delta_epidemic = self.mean_b_epidemic - self.mean_a_epidemic
            if (
                self.warp_mean_a_epidemic.size > 0
                and self.warp_mean_b_epidemic.size > 0
            ):
                self.delta_warp_epidemic = (
                    self.warp_mean_b_epidemic - self.warp_mean_a_epidemic
                )
            else:
                self.delta_warp_epidemic = self.time
        else:
            self.compute_epidemic = False

        return

    def plot(self):
        """
        plot elastic changepoint results

        Usage: obj.plot()
        """

        fig, ax = plt.subplots()
        ax.plot(self.time, self.f, "0.8")
        ax.plot(self.time, self.dat_a, "pink")
        ax.plot(self.time, self.dat_b, "lightskyblue")
        ax.plot(self.time, self.mean_a, "red", label="Before")
        ax.plot(self.time, self.mean_b, "blue", label="After")
        plt.title("Functional Data")
        plt.legend()

        fig, ax = plt.subplots()
        ax.plot(self.time, self.delta, "black")
        plt.title("Estimated Amplitude Change Function")

        fig, ax = plt.subplots()
        ax.plot(self.time, self.warp_data.gam, "0.8")
        ax.plot(self.time, self.warp_a, "pink")
        ax.plot(self.time, self.warp_b, "lightskyblue")
        ax.plot(self.time, self.warp_mean_a, "red", label="Before")
        ax.plot(self.time, self.warp_mean_b, "blue", label="After")
        ax.set_aspect("equal")
        plt.title("Warping Functions")
        plt.legend()

        fig, ax = plt.subplots()
        ax.plot(self.time, self.delta_warp, "black")
        plt.title("Estimated Phase Change Function")

        plt.show()

        if self.compute_epidemic:
            fig, ax = plt.subplots()
            ax.plot(self.time, self.f, "0.8")
            ax.plot(self.time, self.dat_a_epidemic, "pink")
            ax.plot(self.time, self.dat_b_epidemic, "lightskyblue")
            ax.plot(self.time, self.mean_a_epidemic, "red", label="Before/After")
            ax.plot(self.time, self.mean_b_epidemic, "blue", label="During")
            plt.title("Functional Data, Epidemic")
            plt.legend()

            fig, ax = plt.subplots()
            ax.plot(self.time, self.delta_epidemic, "black")
            plt.title("Estimated Amplitude Change Function, Epidemic")

            fig, ax = plt.subplots()
            ax.plot(self.time, self.warp_data.gam, "0.8")
            ax.plot(self.time, self.warp_a_epidemic, "pink")
            ax.plot(self.time, self.warp_b_epidemic, "lightskyblue")
            ax.plot(self.time, self.warp_mean_a, "red", label="Before/After")
            ax.plot(self.time, self.warp_mean_b, "blue", label="During")
            ax.set_aspect("equal")
            plt.title("Warping Functions")
            plt.legend()

            fig, ax = plt.subplots()
            ax.plot(self.time, self.delta_warp_epidemic, "black")
            plt.title("Estimated Phase Change Function, Epidemic")

            plt.show()

            fig, ax = plt.subplots()
            ax.imshow(self.Sn_epidemic)
            plt.title("Test statistic, Epidemic")

            plt.show()

        return


class elastic_amp_change_ff:
    """ "
    This class provides elastic changepoint using elastic FDA.
    It is fully-functional and an extension of the methodology of Aue et al.

    Usage:  obj = elastic_amp_change_ff(f,time)

    :param f: (M,N) % matrix defining N functions of M samples
    :param time: time vector of length M
    :param warp_data: aligned data (default: None)
    :param Sn: test statistic values
    :param Tn: max of test statistic
    :param p: p-value
    :param k_star: change point
    :param values: values of computed Brownian Bridges
    :param dat_a: data before changepoint
    :param dat_b: data after changepoint
    :param warp_a: warping functions before changepoint
    :param warp_b: warping functions after changepoint
    :param mean_a: mean function before changepoint
    :param mean_b: mean function after changepoint
    :param warp_mean_a: mean warping function before changepoint
    :param warp_mean_b: mean warping function after changepoint

    Author :  J. Derek Tucker <jdtuck AT sandia.gov> and Drew Yarger <anyarge AT sandia.gov>
    Date   :  24-Aug-2022

    """

    def __init__(
        self,
        f,
        time,
        smooth_data=False,
        sparam=25,
        use_warp_data=False,
        warp_data=None,
        parallel=False,
    ):
        """
        Construct an instance of the elastic_change class
        :param f: (M,N) % matrix defining N functions of M samples
        :param time: time vector of length M
        :param smooth_data: smooth function data (default: False)
        :param sparam: number of smoothing runs of box filter (default: 25)
        :param use_warp_data: use precomputed warping data (default: False)
        :param warp_data: precomputed aligned data (default: None)
        :param parallel: run computation in parallel (default: True)
        """
        self.f = f
        self.time = time

        if smooth_data:
            self.f = fs.smooth_data(self.f, sparam)

        if use_warp_data:
            self.warp_data = warp_data  # Align Data
        else:
            self.warp_data = fs.fdawarp(self.f, self.time)
            self.warp_data.srsf_align(parallel=parallel)

    def compute(self, d=1000, h=0, M_approx=365, compute_epidemic=False):
        """
        Compute elastic change detection
        :param d: number of monte carlo iterations to compute p-value
        :param h: index of window type to compute long run covariance
        :param M_approx: number of time points to compute p-value
        :param compute_epidemic: compute epidemic changepoint model (default: False)
        """

        M = self.f.shape[0]
        N1 = self.f.shape[1]

        self.mu = np.zeros((M, N1))
        self.mu_f = np.zeros((M, N1))

        # Compute Karcher mean for first i+1 functions
        self.mu[:, 0] = self.warp_data.qn[:, 0]
        self.mu_f[:, 0] = self.warp_data.fn[:, 0]
        for i in range(1, N1):
            self.mu[:, i] = self.warp_data.qn[:, 0 : (i + 1)].sum(axis=1)
            self.mu_f[:, i] = self.warp_data.fn[:, 0 : (i + 1)].sum(axis=1)

        # compute test statistic
        self.Sn = np.zeros(N1 + 1)
        for k in range(1, N1 + 1):
            self.Sn[k] = (
                1
                / M
                * norm(
                    1
                    / np.sqrt(N1)
                    * (self.mu_f[:, k - 1] - (k / N1) * self.mu_f[:, -1])
                )
                ** 2
            )

        self.k_star = np.argmax(self.Sn)
        self.Tn = np.max(self.Sn)

        # compute means on either side of the changepoint
        self.dat_a = self.f[:, range(self.k_star)]
        self.dat_b = self.f[:, range(self.k_star, N1)]
        self.warp_a = self.warp_data.gam[:, 0 : self.k_star]
        self.warp_b = self.warp_data.gam[:, self.k_star :]
        self.mean_a = self.warp_data.fn[:, 0 : self.k_star].mean(axis=1)
        self.mean_b = self.warp_data.fn[:, self.k_star :].mean(axis=1)

        if self.warp_a.size == 0:
            self.warp_mean_a = np.array([])
        else:
            mu, gam_mu, psi, vec = uf.SqrtMean(self.warp_a)
            self.warp_mean_a = gam_mu

        if self.warp_b.size == 0:
            self.warp_mean_b = np.array([])
        else:
            mu, gam_mu, psi, vec = uf.SqrtMean(self.warp_b)
            self.warp_mean_b = gam_mu

        self.delta = self.mean_b - self.mean_a
        if self.warp_mean_a.size > 0 and self.warp_mean_b.size > 0:
            self.delta_warp = self.warp_mean_b - self.warp_mean_a
        else:
            self.delta_warp = self.time

        # center your data
        self.centered_data = np.zeros((M, N1))
        for i in range(self.k_star):
            self.centered_data[:, i] = self.warp_data.fn[:, i] - self.mean_a
        for i in range(self.k_star, N1):
            self.centered_data[:, i] = self.warp_data.fn[:, i] - self.mean_b

        # estimate eigenvalues of covariance operator
        D_mat = LongRunCovMatrixPrecentered(self.centered_data, h=h)
        self.lambda_vals, self.pca_coef = np.linalg.eig(D_mat)
        self.lambda_vals = self.lambda_vals.real / M
        self.pca_coef = self.pca_coef.real

        def asymp(N):
            BridgeLam = np.zeros((M, N))
            for j in range(M):
                BridgeLam[j, :] = self.lambda_vals[j] * (
                    BBridge(0, 0, 0, 1, N - 1) ** 2
                )
            return max(BridgeLam.sum(axis=0))

        values = np.zeros(d)
        for sim in range(d):
            values[sim] = asymp(N1)

        z = self.Tn <= values
        self.p = z.mean()
        self.values = values

        if compute_epidemic:
            self.compute_epidemic = True
            snk_minus_vals = np.zeros((N1, N1, M))
            snk_minus_summary = np.zeros((N1, N1))
            for i in range(N1):
                for j in range(N1):
                    if i > j:
                        continue
                    snk_minus_vals[i, j, :] = (
                        1
                        / np.sqrt(N1)
                        * (
                            self.warp_data.fn[:, i : (j + 1)].sum(axis=1)
                            - ((j - i + 1) / N1) * self.mu_f[:, -1]
                        )
                    )
                    snk_minus_summary[i, j] = (
                        1 / M * (norm(snk_minus_vals[i, j, :]) ** 2)
                    )
            self.tau1 = np.argmax(snk_minus_summary.max(axis=1))
            self.tau2 = np.argmax(snk_minus_summary[self.tau1, :])
            self.Sn_epidemic = snk_minus_summary
            self.Tn_epidemic = snk_minus_summary.max()

            # compute means on either side of the changepoint
            self.dat_a_epidemic = self.f[:, range(self.tau1, self.tau2 + 1)]
            self.dat_b_epidemic = np.column_stack(
                (self.f[:, range(0, self.tau1)], self.f[:, range(self.tau2 + 1, N1)])
            )
            self.warp_a_epidemic = self.warp_data.gam[
                :, range(self.tau1, self.tau2 + 1)
            ]
            self.warp_b_epidemic = np.column_stack(
                (
                    self.warp_data.gam[:, range(0, self.tau1)],
                    self.warp_data.gam[:, range(self.tau2 + 1, N1)],
                )
            )
            self.mean_a_epidemic = self.warp_data.fn[
                :, range(self.tau1, self.tau2 + 1)
            ].mean(axis=1)
            self.mean_b_epidemic = np.column_stack(
                (
                    self.warp_data.fn[:, range(0, self.tau1)],
                    self.warp_data.fn[:, range(self.tau2 + 1, N1)],
                )
            ).mean(axis=1)

            if self.warp_a_epidemic.size == 0:
                self.warp_mean_a_epidemic = np.array([])
            else:
                mu, gam_mu, psi, vec = uf.SqrtMean(self.warp_a_epidemic)
                self.warp_mean_a_epidemic = gam_mu

            if self.warp_b_epidemic.size == 0:
                self.warp_mean_b_epidemic = np.array([])
            else:
                mu, gam_mu, psi, vec = uf.SqrtMean(self.warp_b_epidemic)
                self.warp_mean_b_epidemic = gam_mu

            self.delta_epidemic = self.mean_b_epidemic - self.mean_a_epidemic
            if (
                self.warp_mean_a_epidemic.size > 0
                and self.warp_mean_b_epidemic.size > 0
            ):
                self.delta_warp_epidemic = (
                    self.warp_mean_b_epidemic - self.warp_mean_a_epidemic
                )
            else:
                self.delta_warp_epidemic = self.time

            # center your data
            self.centered_data_epidemic = np.zeros((M, N1))
            for i in range(N1):
                if i in range(self.tau1, self.tau2 + 1):
                    self.centered_data_epidemic[:, i] = (
                        self.warp_data.fn[:, i] - self.mean_b
                    )
                else:
                    self.centered_data_epidemic[:, i] = (
                        self.warp_data.fn[:, i] - self.mean_a
                    )
                    # estimate eigenvalues of covariance operator
            D_mat = LongRunCovMatrixPrecentered(self.centered_data_epidemic, h=h)
            self.lambda_vals_epidemic, self.pca_coef_epidemic = np.linalg.eig(D_mat)
            self.lambda_vals_epidemic = self.lambda_vals_epidemic.real / M
            self.pca_coef_epidemic = self.pca_coef_epidemic.real * M

            def asymp(N):
                BridgeLam = np.zeros((M_approx, N, N))
                for j in range(M_approx):
                    BB = BBridge(0, 0, 0, 1, N - 1)
                    BB_sub = np.broadcast_to(BB, (N, N)) - np.broadcast_to(BB, (N, N)).T
                    BridgeLam[j, :, :] = self.lambda_vals_epidemic[j] * (BB_sub**2)
                return BridgeLam.sum(axis=0).max()

            values = np.zeros(d)
            for sim in range(d):
                values[sim] = asymp(N1)

            z = self.Tn_epidemic <= values
            self.p_epidemic = z.mean()
            self.values_epidemic = values
        else:
            self.compute_epidemic = False

        return

    def plot(self):
        """
        plot elastic changepoint results

        Usage: obj.plot()
        """

        fig, ax = plt.subplots()
        ax.plot(self.time, self.f, "0.8")
        ax.plot(self.time, self.dat_a, "pink")
        ax.plot(self.time, self.dat_b, "lightskyblue")
        ax.plot(self.time, self.mean_a, "red", label="Before")
        ax.plot(self.time, self.mean_b, "blue", label="After")
        plt.title("Functional Data")
        plt.legend()

        fig, ax = plt.subplots()
        ax.plot(self.time, self.delta, "black")
        plt.title("Estimated Amplitude Change Function")

        fig, ax = plt.subplots()
        ax.plot(self.time, self.warp_data.gam, "0.8")
        ax.plot(self.time, self.warp_data.gam[:, range(self.k_star)], "pink")
        ax.plot(self.time, self.warp_data.gam[:, self.k_star :], "lightskyblue")
        ax.plot(self.time, self.warp_mean_a, "red", label="Before")
        ax.plot(self.time, self.warp_mean_b, "blue", label="After")
        ax.set_aspect("equal")
        plt.title("Warping Functions")
        plt.legend()

        fig, ax = plt.subplots()
        ax.plot(list(range(self.Sn.shape[0])), self.Sn)
        plt.title("Test statistic")

        plt.show()

        if self.compute_epidemic:
            fig, ax = plt.subplots()
            ax.plot(self.time, self.f, "0.8")
            ax.plot(self.time, self.dat_a_epidemic, "pink")
            ax.plot(self.time, self.dat_b_epidemic, "lightskyblue")
            ax.plot(self.time, self.mean_a_epidemic, "red", label="Before/After")
            ax.plot(self.time, self.mean_b_epidemic, "blue", label="During")
            plt.title("Functional Data, Epidemic")
            plt.legend()

            fig, ax = plt.subplots()
            ax.plot(self.time, self.delta_epidemic, "black")
            plt.title("Estimated Amplitude Change Function, Epidemic")

            fig, ax = plt.subplots()
            ax.plot(self.time, self.warp_data.gam, "0.8")
            ax.plot(self.time, self.warp_a_epidemic, "pink")
            ax.plot(self.time, self.warp_b_epidemic, "lightskyblue")
            ax.plot(self.time, self.warp_mean_a, "red", label="Before/After")
            ax.plot(self.time, self.warp_mean_b, "blue", label="During")
            ax.set_aspect("equal")
            plt.title("Warping Functions")
            plt.legend()

            fig, ax = plt.subplots()
            ax.plot(self.time, self.delta_warp_epidemic, "black")
            plt.title("Estimated Phase Change Function, Epidemic")

            plt.show()

            fig, ax = plt.subplots()
            ax.imshow(self.Sn_epidemic)
            plt.title("Test statistic, Epidemic")

            plt.show()

        return


class elastic_ph_change_ff:
    """ "
    This class provides elastic changepoint using elastic FDA on warping functions.
    It is fully-functional and an extension of the methodology of Aue et al.

    Usage:  obj = elastic_ph_change_ff(f,time)

    :param f: (M,N) % matrix defining N functions of M samples
    :param time: time vector of length M
    :param warp_data: aligned data (default: None)
    :param Sn: test statistic values
    :param Tn: max of test statistic
    :param p: p-value
    :param k_star: change point
    :param values: values of computed Brownian Bridges
    :param dat_a: data before changepoint
    :param dat_b: data after changepoint
    :param warp_a: warping functions before changepoint
    :param warp_b: warping functions after changepoint
    :param mean_a: mean function before changepoint
    :param mean_b: mean function after changepoint
    :param warp_mean_a: mean warping function before changepoint
    :param warp_mean_b: mean warping function after changepoint

    Author :  J. Derek Tucker <jdtuck AT sandia.gov>
    Date   :  17-Nov-2022

    """

    def __init__(
        self,
        f,
        time,
        smooth_data=False,
        sparam=25,
        use_warp_data=False,
        warp_data=None,
        parallel=False,
    ):
        """
        Construct an instance of the elastic_change class
        :param f: (M,N) % matrix defining N functions of M samples
        :param time: time vector of length M
        :param smooth_data: smooth function data (default: False)
        :param sparam: number of smoothing runs of box filter (default: 25)
        :param use_warp_data: use precomputed warping data (default: False)
        :param warp_data: precomputed aligned data (default: None)
        :param parallel: run computation in parallel (default: True)
        """

        self.f = f
        self.time = time

        if smooth_data:
            self.f = fs.smooth_data(self.f, sparam)

        if use_warp_data:
            self.warp_data = warp_data  # Align Data
        else:
            self.warp_data = fs.fdawarp(self.f, self.time)
            self.warp_data.srsf_align(parallel=parallel)

    def compute(self, d=1000, h=0, M_approx=365):
        """
        Compute elastic change detection
        :param d: number of monte carlo iterations to compute p-value
        :param h: index of window type to compute long run covariance
        :param M_approx: number of time points to compute p-value
        """

        M = self.f.shape[0]
        N1 = self.f.shape[1]

        self.mu = np.zeros((M, N1))

        # Compute Karcher mean of warping functions
        mu, gam_mu, psi, vec = uf.SqrtMean(self.warp_data.gam)
        self.vec = vec

        self.mu[:, 0] = vec[:, 0]
        for i in range(1, N1):
            self.mu[:, i] = vec[:, 0 : (i + 1)].sum(axis=1)

        # compute test statistic
        self.Sn = np.zeros(N1 + 1)
        # k indexes subscript of test statistic
        for k in range(1, N1 + 1):
            self.Sn[k] = (
                norm(1 / np.sqrt(N1) * (self.mu[:, k - 1] - (k / N1) * self.mu[:, -1]))
                ** 2
                / M
            )

        self.k_star = np.argmax(self.Sn)
        self.Tn = np.max(self.Sn)

        # compute means on either side of the changepoint
        self.dat_a = self.f[:, range(self.k_star)]
        self.dat_b = self.f[:, range(self.k_star, N1)]
        self.warp_a = self.warp_data.gam[:, 0 : self.k_star]
        self.warp_b = self.warp_data.gam[:, self.k_star :]
        self.mean_a = self.warp_data.fn[:, 0 : self.k_star].mean(axis=1)
        self.mean_b = self.warp_data.fn[:, self.k_star :].mean(axis=1)

        if self.warp_a.size == 0:
            mu_a = np.zeros(M)
            self.warp_mean_a = np.array([])
        else:
            mu, gam_mu, psi, vec = uf.SqrtMean(self.warp_a)
            mu_a = vec.mean(axis=1)
            self.warp_mean_a = gam_mu

        if self.warp_b.size == 0:
            mu_b = np.zeros(M)
            self.warp_mean_b = np.array([])
        else:
            mu, gam_mu, psi, vec = uf.SqrtMean(self.warp_b)
            mu_b = vec.mean(axis=1)
            self.warp_mean_b = gam_mu

        self.delta = self.mean_b - self.mean_a
        if self.warp_mean_a.size > 0 and self.warp_mean_b.size > 0:
            self.delta_warp = self.warp_mean_b - self.warp_mean_a
        else:
            self.delta_warp = self.time

        # center your data
        self.centered_data = np.zeros((M, N1))
        for i in range(self.k_star):
            self.centered_data[:, i] = self.vec[:, i] - mu_a
        for i in range(self.k_star, N1):
            self.centered_data[:, i] = self.vec[:, i] - mu_b

        # estimate eigenvalues of covariance operator
        D_mat = LongRunCovMatrixPrecentered(self.centered_data, h=h)
        self.lambda_vals, self.pca_coef = np.linalg.eig(D_mat)
        self.lambda_vals = self.lambda_vals.real / M
        self.pca_coef = self.pca_coef.real * M

        def asymp(N):
            BridgeLam = np.zeros((M, N))
            for j in range(M):
                BridgeLam[j, :] = self.lambda_vals[j] * (
                    BBridge(0, 0, 0, 1, N - 1) ** 2
                )
            return max(BridgeLam.sum(axis=0))

        values = np.zeros(d)
        for sim in range(d):
            values[sim] = asymp(N1)

        z = self.Tn <= values
        self.p = z.mean()
        self.values = values

        return

    def plot(self):
        """
        plot elastic changepoint results

        Usage: obj.plot()
        """

        fig, ax = plt.subplots()
        ax.plot(self.time, self.f, "0.8")
        ax.plot(self.time, self.dat_a, "pink")
        ax.plot(self.time, self.dat_b, "lightskyblue")
        ax.plot(self.time, self.mean_a, "red", label="Before")
        ax.plot(self.time, self.mean_b, "blue", label="After")
        plt.title("Functional Data")
        plt.legend()

        fig, ax = plt.subplots()
        ax.plot(self.time, self.delta, "black")
        plt.title("Estimated Amplitude Change Function")

        fig, ax = plt.subplots()
        ax.plot(self.time, self.warp_data.gam, "0.8")
        ax.plot(self.time, self.warp_data.gam[:, range(self.k_star)], "pink")
        ax.plot(self.time, self.warp_data.gam[:, self.k_star :], "lightskyblue")
        ax.plot(self.time, self.warp_mean_a, "red", label="Before")
        ax.plot(self.time, self.warp_mean_b, "blue", label="After")
        ax.set_aspect("equal")
        plt.title("Warping Functions")
        plt.legend()

        fig, ax = plt.subplots()
        ax.plot(list(range(self.Sn.shape[0])), self.Sn)
        plt.title("Test statistic")

        plt.show()

        return



"""
Vertical and Horizontal Functional Principal Component Analysis using SRSF

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""
import numpy as np
import fdasrsf as fs
import fdasrsf.utility_functions as uf
import fdasrsf.geometry as geo
from scipy.linalg import norm, svd
from scipy.integrate import trapezoid, cumulative_trapezoid
from scipy.optimize import fminbound
from joblib import Parallel, delayed
import matplotlib.pyplot as plt
import fdasrsf.plot_style as plot


class fdavpca:
    """
    This class provides vertical fPCA using the
    SRVF framework

    Usage:  obj = fdavpca(warp_data)

    :param warp_data: fdawarp class with alignment data
    :param q_pca: srvf principal directions
    :param f_pca: f principal directions
    :param latent: latent values
    :param coef: principal coefficients
    :param id: point used for f(0)
    :param mqn: mean srvf
    :param U: eigenvectors
    :param stds: geodesic directions
    :param new_coef: principal coefficients of new data

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  15-Mar-2018
    """

    def __init__(self, fdawarp):
        """
        Construct an instance of the fdavpca class
        :param fdawarp: fdawarp class
        """
        if fdawarp.fn.size == 0:
            raise Exception("Please align fdawarp class using srsf_align!")

        self.warp_data = fdawarp

    def calc_fpca(self, no=3, var_exp=None, id=None, stds=np.arange(-1, 2)):
        """
        This function calculates vertical functional principal component
        analysis on aligned data

        :param no: number of components to extract (default = 3)
        :param var_exp: compute no based on value percent variance explained
                        (example: 0.95)
        :param id: point to use for f(0) (default = midpoint)
        :param stds: number of standard deviations along geodesic to compute
                     (default = -1,0,1)
        :type no: int
        :type id: int

        :rtype: fdavpca object containing
        :return q_pca: srsf principal directions
        :return f_pca: functional principal directions
        :return latent: latent values
        :return coef: coefficients
        :return U: eigenvectors

        """
        fn = self.warp_data.fn
        time = self.warp_data.time
        qn = self.warp_data.qn

        M = time.shape[0]
        if var_exp is not None:
            if var_exp > 1:
                raise Exception("var_exp is greater than 1")
            no = M

        if id is None:
            mididx = int(np.round(M / 2))
        else:
            mididx = id

        Nstd = stds.shape[0]

        # FPCA
        mq_new = qn.mean(axis=1)
        N = mq_new.shape[0]
        m_new = np.sign(fn[mididx, :]) * np.sqrt(np.abs(fn[mididx, :]))
        mqn = np.append(mq_new, m_new.mean())
        self.mqn2 = mqn
        qn2 = np.vstack((qn, m_new))
        K = np.cov(qn2)

        U, s, V = svd(K)
        stdS = np.sqrt(s)

        # compute the PCA in the q domain
        q_pca = np.ndarray(shape=(N + 1, Nstd, no), dtype=float)
        for k in range(0, no):
            for l in range(0, Nstd):
                q_pca[:, l, k] = mqn + stds[l] * stdS[k] * U[:, k]

        # compute the correspondence in the f domain
        f_pca = np.ndarray(shape=(N, Nstd, no), dtype=float)
        for k in range(0, no):
            for l in range(0, Nstd):
                f_pca[:, l, k] = uf.cumtrapzmid(
                    time,
                    q_pca[0:N, l, k] * np.abs(q_pca[0:N, l, k]),
                    np.sign(q_pca[N, l, k]) * (q_pca[N, l, k] ** 2),
                    mididx,
                )
            fbar = fn.mean(axis=1)
            fsbar = f_pca[:, :, k].mean(axis=1)
            err = np.transpose(np.tile(fbar - fsbar, (Nstd, 1)))
            f_pca[:, :, k] += err

        N2 = qn.shape[1]
        c = np.zeros((N2, no))
        for k in range(0, no):
            for l in range(0, N2):
                c[l, k] = sum((np.append(qn[:, l], m_new[l]) - mqn) * U[:, k])

        if var_exp is not None:
            cumm_coef = np.cumsum(s) / sum(s)
            no = int(np.argwhere(cumm_coef <= var_exp)[-1])

        self.q_pca = q_pca
        self.f_pca = f_pca
        self.latent = s[0:no]
        self.coef = c[:, 0:no]
        self.U = U[:, 0:no]
        self.id = mididx
        self.mqn = mqn
        self.time = time
        self.stds = stds
        self.no = no

        return

    def project(self, f):
        """
        project new data onto fPCA basis

        Usage: obj.project(f)

        :param f: numpy array (MxN) of N functions on M time

        """

        q1 = fs.f_to_srsf(f, self.time)
        M = self.time.shape[0]
        n = q1.shape[1]
        mq = self.warp_data.mqn
        fn = np.zeros((M, n))
        qn = np.zeros((M, n))
        gam = np.zeros((M, n))
        for ii in range(0, n):
            gam[:, ii] = fs.optimum_reparam(mq, self.time, q1[:, ii])
            fn[:, ii] = fs.warp_f_gamma(self.time, f[:, ii], gam[:, ii])
            qn[:, ii] = fs.f_to_srsf(fn[:, ii], self.time)

        U = self.U
        no = U.shape[1]

        m_new = np.sign(fn[self.id, :]) * np.sqrt(np.abs(fn[self.id, :]))
        qn1 = np.vstack((qn, m_new))

        a = np.zeros((n, no))
        for i in range(0, n):
            for j in range(0, no):
                a[i, j] = sum((qn1[:, i] - self.mqn2) * U[:, j])

        self.new_coef = a

        return

    def plot(self):
        """
        plot plot elastic vertical fPCA result
        Usage: obj.plot()
        """

        no = self.no
        Nstd = self.stds.shape[0]
        N = self.time.shape[0]
        num_plot = int(np.ceil(no / 3))
        CBcdict = {
            "Bl": (0, 0, 0),
            "Or": (0.9, 0.6, 0),
            "SB": (0.35, 0.7, 0.9),
            "bG": (0, 0.6, 0.5),
            "Ye": (0.95, 0.9, 0.25),
            "Bu": (0, 0.45, 0.7),
            "Ve": (0.8, 0.4, 0),
            "rP": (0.8, 0.6, 0.7),
        }
        cl = sorted(CBcdict.keys())
        k = 0
        for ii in range(0, num_plot):
            if k > (no - 1):
                break

            fig, ax = plt.subplots(2, 3)

            for k1 in range(0, 3):
                k = k1 + (ii) * 3
                axt = ax[0, k1]
                if k > (no - 1):
                    break

                for l in range(0, Nstd):
                    axt.plot(self.time, self.q_pca[0:N, l, k], color=CBcdict[cl[l]])

                axt.set_title("q domain: PD %d" % (k + 1))

                axt = ax[1, k1]
                for l in range(0, Nstd):
                    axt.plot(self.time, self.f_pca[:, l, k], color=CBcdict[cl[l]])

                axt.set_title("f domain: PD %d" % (k + 1))

            fig.set_tight_layout(True)

        cumm_coef = 100 * np.cumsum(self.latent) / sum(self.latent)
        N = self.latent.shape[0]
        idx = np.arange(0, N) + 1
        plot.f_plot(idx, cumm_coef, "Coefficient Cumulative Percentage")
        plt.ylabel("Percentage")
        plt.xlabel("Index")
        plt.show()

        return


class fdahpca:
    """
    This class provides horizontal fPCA using the
    SRVF framework

    Usage:  obj = fdahpca(warp_data)

    :param warp_data: fdawarp class with alignment data
    :param gam_pca: warping functions principal directions
    :param psi_pca: srvf principal directions
    :param latent: latent values
    :param U: eigenvectors
    :param coef: coefficients
    :param vec: shooting vectors
    :param mu: Karcher Mean
    :param tau: principal directions

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  15-Mar-2018
    """

    def __init__(self, fdawarp):
        """
        Construct an instance of the fdavpca class
        :param fdawarp: fdawarp class
        """
        if fdawarp.fn.size == 0:
            raise Exception("Please align fdawarp class using srsf_align!")

        self.warp_data = fdawarp

    def calc_fpca(self, no=3, var_exp=None, stds=np.arange(-1, 2)):
        """
        This function calculates horizontal functional principal component
        analysis on aligned data

        :param no: number of components to extract (default = 3)
        :param var_exp: compute no based on value percent variance explained
                        (example: 0.95)
        :param stds: number of standard deviations along geodesic to compute
                     (default = -1,0,1)
        :type no: int

        :rtype: fdahpca object of numpy ndarray
        :return q_pca: srsf principal directions
        :return f_pca: functional principal directions
        :return latent: latent values
        :return coef: coefficients
        :return U: eigenvectors

        """
        # Calculate Shooting Vectors
        gam = self.warp_data.gam
        mu, gam_mu, psi, vec = uf.SqrtMean(gam)
        TT = self.warp_data.time.shape[0]

        if var_exp is not None:
            if var_exp > 1:
                raise Exception("var_exp is greater than 1")
            no = TT

        # TFPCA
        K = np.cov(vec)

        U, s, V = svd(K)
        vm = vec.mean(axis=1)
        self.vm = vm
        self.mu_psi = mu

        gam_pca = np.ndarray(shape=(stds.shape[0], mu.shape[0], no), dtype=float)
        psi_pca = np.ndarray(shape=(stds.shape[0], mu.shape[0], no), dtype=float)
        for j in range(0, no):
            cnt = 0
            for k in stds:
                v = k * np.sqrt(s[j]) * U[:, j]
                vn = norm(v) / np.sqrt(TT)
                if vn < 0.0001:
                    psi_pca[cnt, :, j] = mu
                else:
                    psi_pca[cnt, :, j] = np.cos(vn) * mu + np.sin(vn) * v / vn

                tmp = cumulative_trapezoid(
                    psi_pca[cnt, :, j] * psi_pca[cnt, :, j],
                    np.linspace(0, 1, TT),
                    initial=0,
                )
                gam_pca[cnt, :, j] = (tmp - tmp[0]) / (tmp[-1] - tmp[0])
                cnt += 1

        N2 = gam.shape[1]
        c = np.zeros((N2, no))
        for k in range(0, no):
            for i in range(0, N2):
                c[i, k] = np.dot(vec[:, i] - vm, U[:, k])

        if var_exp is not None:
            cumm_coef = np.cumsum(s) / sum(s)
            no = int(np.argwhere(cumm_coef <= var_exp)[-1])

        self.gam_pca = gam_pca
        self.psi_pca = psi_pca
        self.U = U[:, 0:no]
        self.coef = c[:, 0:no]
        self.latent = s[0:no]
        self.gam_mu = gam_mu
        self.psi_mu = mu
        self.vec = vec
        self.no = no
        self.stds = stds

        return

    def project(self, f):
        """
        project new data onto fPCA basis

        Usage: obj.project(f)

        :param f: numpy array (MxN) of N functions on M time

        """

        q1 = fs.f_to_srsf(f, self.time)
        M = self.time.shape[0]
        n = q1.shape[1]
        mq = self.warp_data.mqn
        gam = np.zeros((M, n))
        for ii in range(0, n):
            gam[:, ii] = fs.optimum_reparam(mq, self.time, q1[:, ii])

        U = self.U
        no = U.shape[1]

        mu_psi = self.mu_psi
        vec = np.zeros((M, n))
        psi = np.zeros((M, n))
        time = np.linspace(0, 1, M)
        binsize = np.mean(np.diff(time))
        for i in range(0, n):
            psi[:, i] = np.sqrt(np.gradient(gam[:, i], binsize))
            out, theta = fs.inv_exp_map(mu_psi, psi[:, i])
            vec[:, i] = out

        a = np.zeros((n, no))
        for i in range(0, n):
            for j in range(0, no):
                a[i, j] = np.dot(vec[:, i] - self.vm, U[:, j])

        self.new_coef = a

        return

    def plot(self):
        """
        plot plot elastic horizontal fPCA results

        Usage: obj.plot()
        """

        no = self.no
        TT = self.warp_data.time.shape[0]
        num_plot = int(np.ceil(no / 3))
        CBcdict = {
            "Bl": (0, 0, 0),
            "Or": (0.9, 0.6, 0),
            "SB": (0.35, 0.7, 0.9),
            "bG": (0, 0.6, 0.5),
            "Ye": (0.95, 0.9, 0.25),
            "Bu": (0, 0.45, 0.7),
            "Ve": (0.8, 0.4, 0),
            "rP": (0.8, 0.6, 0.7),
        }
        k = 0
        for ii in range(0, num_plot):
            if k > (no - 1):
                break

            fig, ax = plt.subplots(1, 3)

            for k1 in range(0, 3):
                k = k1 + (ii) * 3
                axt = ax[k1]
                if k > (no - 1):
                    break

                tmp = self.gam_pca[:, :, k]
                axt.plot(np.linspace(0, 1, TT), tmp.transpose())
                axt.set_title("PD %d" % (k + 1))
                axt.set_aspect("equal")

            fig.set_tight_layout(True)

        cumm_coef = 100 * np.cumsum(self.latent[0:no]) / sum(self.latent[0:no])
        idx = np.arange(0, no) + 1
        plot.f_plot(idx, cumm_coef, "Coefficient Cumulative Percentage")
        plt.ylabel("Percentage")
        plt.xlabel("Index")
        plt.show()

        return


class fdajpca:
    """
    This class provides joint fPCA using the
    SRVF framework

    Usage:  obj = fdajpca(warp_data)

    :param warp_data: fdawarp class with alignment data
    :param q_pca: srvf principal directions
    :param f_pca: f principal directions
    :param latent: latent values
    :param coef: principal coefficients
    :param id: point used for f(0)
    :param mqn: mean srvf
    :param U: eigenvectors
    :param mu_psi: mean psi
    :param mu_g: mean g
    :param C: scaling value
    :param stds: geodesic directions

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  18-Mar-2018
    """

    def __init__(self, fdawarp):
        """
        Construct an instance of the fdavpca class
        :param fdawarp: fdawarp class
        """
        if fdawarp.fn.size == 0:
            raise Exception("Please align fdawarp class using srsf_align!")

        self.warp_data = fdawarp
        self.M = fdawarp.time.shape[0]

    def calc_fpca(
        self,
        no=3,
        var_exp=None,
        stds=np.arange(-1.0, 2.0),
        id=None,
        parallel=False,
        cores=-1,
    ):
        """
        This function calculates joint functional principal component analysis
        on aligned data

        :param no: number of components to extract (default = 3)
        :param var_exp: compute no based on value percent variance explained
                        (example: 0.95)
        :param id: point to use for f(0) (default = midpoint)
        :param stds: number of standard deviations along gedoesic to compute
                     (default = -1,0,1)
        :param parallel: run in parallel (default = F)
        :param cores: number of cores for parallel (default = -1 (all))
        :type no: int
        :type id: int
        :type parallel: bool
        :type cores: int

        :rtype: fdajpca object of numpy ndarray
        :return q_pca: srsf principal directions
        :return f_pca: functional principal directions
        :return latent: latent values
        :return coef: coefficients
        :return U: eigenvectors

        """
        fn = self.warp_data.fn
        time = self.warp_data.time
        qn = self.warp_data.qn
        q0 = self.warp_data.q0
        gam = self.warp_data.gam

        M = time.shape[0]
        if var_exp is not None:
            if var_exp > 1:
                raise Exception("var_exp is greater than 1")
            no = M

        if id is None:
            mididx = int(np.round(M / 2))
        else:
            mididx = id

        Nstd = stds.shape[0]

        # set up for fPCA in q-space
        mq_new = qn.mean(axis=1)
        m_new = np.sign(fn[mididx, :]) * np.sqrt(np.abs(fn[mididx, :]))
        mqn = np.append(mq_new, m_new.mean())
        qn2 = np.vstack((qn, m_new))

        # calculate vector space of warping functions
        mu_psi, gam_mu, psi, vec = uf.SqrtMean(gam, parallel, cores)

        # joint fPCA
        C = fminbound(find_C, 0, 1e4, (qn2, vec, q0, no, mu_psi, parallel, cores))
        qhat, gamhat, a, U, s, mu_g, g, cov = jointfPCAd(
            qn2, vec, C, no, mu_psi, parallel, cores
        )

        # geodesic paths
        q_pca = np.ndarray(shape=(M, Nstd, no), dtype=float)
        f_pca = np.ndarray(shape=(M, Nstd, no), dtype=float)

        for k in range(0, no):
            for l in range(0, Nstd):
                qhat = mqn + np.dot(U[0: (M + 1), k], stds[l] * np.sqrt(s[k]))
                vechat = np.dot(U[(M + 1):, k], (stds[l] * np.sqrt(s[k])) / C)
                psihat = geo.exp_map(mu_psi, vechat)
                gamhat = cumulative_trapezoid(psihat * psihat,
                                              np.linspace(0, 1, M),
                                              initial=0)
                gamhat = (gamhat - gamhat.min()) / (gamhat.max() - gamhat.min())
                if sum(vechat) == 0:
                    gamhat = np.linspace(0, 1, M)

                fhat = uf.cumtrapzmid(
                    time,
                    qhat[0:M] * np.fabs(qhat[0:M]),
                    np.sign(qhat[M]) * (qhat[M] * qhat[M]),
                    mididx,
                )
                f_pca[:, l, k] = uf.warp_f_gamma(np.linspace(0, 1, M), 
                                                 fhat, gamhat)
                q_pca[:, l, k] = uf.warp_q_gamma(
                    np.linspace(0, 1, M), qhat[0:M], gamhat
                )

        if var_exp is not None:
            cumm_coef = np.cumsum(s) / s.sum()
            no = int(np.argwhere(cumm_coef >= var_exp)[0][0])

        self.q_pca = q_pca
        self.f_pca = f_pca
        self.latent = s[0:no]
        self.coef = a[:, 0:no]
        self.U = U[:, 0:no]
        self.mu_psi = mu_psi
        self.mu_g = mu_g
        self.id = mididx
        self.C = C
        self.time = time
        self.g = g
        self.cov = cov
        self.no = no
        self.stds = stds

        return

    def project(self, f):
        """
        project new data onto fPCA basis

        Usage: obj.project(f)

        :param f: numpy array (MxN) of N functions on M time

        """

        q1 = fs.f_to_srsf(f, self.time)
        M = self.time.shape[0]
        n = q1.shape[1]
        mq = self.warp_data.mqn
        fn = np.zeros((M, n))
        qn = np.zeros((M, n))
        gam = np.zeros((M, n))
        for ii in range(0, n):
            gam[:, ii] = fs.optimum_reparam(mq, self.time, q1[:, ii])
            fn[:, ii] = fs.warp_f_gamma(self.time, f[:, ii], gam[:, ii])
            qn[:, ii] = fs.f_to_srsf(fn[:, ii], self.time)

        U = self.U
        no = U.shape[1]

        m_new = np.sign(fn[self.id, :]) * np.sqrt(np.abs(fn[self.id, :]))
        qn1 = np.vstack((qn, m_new))
        C = self.C
        TT = self.time.shape[0]
        mu_g = self.mu_g
        mu_psi = self.mu_psi
        vec = np.zeros((M, n))
        psi = np.zeros((TT, n))
        time = np.linspace(0, 1, TT)
        binsize = np.mean(np.diff(time))
        for i in range(0, n):
            psi[:, i] = np.sqrt(np.gradient(gam[:, i], binsize))
            out, theta = fs.inv_exp_map(mu_psi, psi[:, i])
            vec[:, i] = out

        g = np.vstack((qn1, C * vec))
        a = np.zeros((n, no))
        for i in range(0, n):
            for j in range(0, no):
                tmp = g[:, i] - mu_g
                a[i, j] = np.dot(tmp.T, U[:, j])

        self.new_g = g
        self.new_coef = a

        return

    def plot(self):
        """
        plot plot elastic joint fPCA result

        Usage: obj.plot()
        """
        no = self.no
        M = self.time.shape[0]
        Nstd = self.stds.shape[0]
        num_plot = int(np.ceil(no / 3))
        CBcdict = {
            "Bl": (0, 0, 0),
            "Or": (0.9, 0.6, 0),
            "SB": (0.35, 0.7, 0.9),
            "bG": (0, 0.6, 0.5),
            "Ye": (0.95, 0.9, 0.25),
            "Bu": (0, 0.45, 0.7),
            "Ve": (0.8, 0.4, 0),
            "rP": (0.8, 0.6, 0.7),
        }
        cl = sorted(CBcdict.keys())
        k = 0
        for ii in range(0, num_plot):
            if k > (no - 1):
                break

            fig, ax = plt.subplots(2, 3)

            for k1 in range(0, 3):
                k = k1 + (ii) * 3
                axt = ax[0, k1]
                if k > (no - 1):
                    break

                for l in range(0, Nstd):
                    axt.plot(self.time, self.q_pca[0:M, l, k], color=CBcdict[cl[l]])

                axt.set_title("q domain: PD %d" % (k + 1))

                axt = ax[1, k1]
                for l in range(0, Nstd):
                    axt.plot(self.time, self.f_pca[:, l, k], color=CBcdict[cl[l]])

                axt.set_title("f domain: PD %d" % (k + 1))

            fig.set_tight_layout(True)

        cumm_coef = 100 * np.cumsum(self.latent) / sum(self.latent)
        idx = np.arange(0, self.latent.shape[0]) + 1
        plot.f_plot(idx, cumm_coef, "Coefficient Cumulative Percentage")
        plt.ylabel("Percentage")
        plt.xlabel("Index")
        plt.show()

        return


class fdajpcah:
    """
    This class provides joint fPCA using the
    SRVF framework using MFPCA

    Usage:  obj = fdajpcah(warp_data)

    :param warp_data: fdawarp class with alignment data
    :param q_pca: srvf principal directions
    :param f_pca: f principal directions
    :param latent: latent values
    :param coef: principal coefficients
    :param id: point used for f(0)
    :param mqn: mean srvf
    :param U_q: eigenvectors for q
    :param U_h: eigenvectors for gam
    :param C: scaling value
    :param stds: geodesic directions

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  06-April-2024
    """

    def __init__(self, fdawarp):
        """
        Construct an instance of the fdavpca class
        :param fdawarp: fdawarp class
        """
        if fdawarp.fn.size == 0:
            raise Exception("Please align fdawarp class using srsf_align!")

        self.warp_data = fdawarp
        self.M = fdawarp.time.shape[0]

    def calc_fpca(
        self,
        var_exp=0.99,
        stds=np.arange(-1.0, 2.0),
        id=None,
        parallel=False,
        cores=-1,
        srsf=True,
    ):
        """
        This function calculates joint functional principal component analysis
        on aligned data

        :param var_exp: compute no based on value percent variance explained
                        (default: None)
        :param id: point to use for f(0) (default = midpoint)
        :param stds: number of standard deviations along gedoesic to compute
                     (default = -1,0,1)
        :param parallel: run in parallel (default = F)
        :param cores: number of cores for parallel (default = -1 (all))
        :type id: int
        :type parallel: bool
        :type cores: int

        :rtype: fdajpcah object of numpy ndarray
        :return q_pca: srsf principal directions
        :return f_pca: functional principal directions
        :return latent: latent values
        :return coef: coefficients
        :param U_q: eigenvectors for q
        :param U_h: eigenvectors for gam

        """
        fn = self.warp_data.fn
        time = self.warp_data.time
        qn = self.warp_data.qn
        gam = self.warp_data.gam

        M = time.shape[0]
        if var_exp is not None:
            if var_exp > 1:
                raise Exception("var_exp is greater than 1")
            no = M

        if id is None:
            mididx = int(np.round(M / 2))
        else:
            mididx = id

        Nstd = stds.shape[0]

        # set up for fPCA in q-space
        if srsf:
            mq_new = qn.mean(axis=1)
            m_new = np.sign(fn[mididx, :]) * np.sqrt(np.abs(fn[mididx, :]))
            mqn = np.append(mq_new, m_new.mean())
            qn2 = np.vstack((qn, m_new))
            q0 = self.warp_data.q0
        else:
            mqn = fn.mean(axis=1)
            q0 = self.warp_data.f
            qn2 = fn.copy()

        # calculate vector space of warping functions
        h = geo.gam_to_h(gam)

        # joint fPCA
        C = fminbound(find_C_h, 0, 200, (qn2, h, q0, 0.99, parallel, cores, srsf))
        qhat, gamhat, cz, Psi_q, Psi_h, sz, U, Uh, Uz = jointfPCAhd(qn2, h, C, var_exp)

        hc = C * h
        mh = np.mean(hc, axis=1)

        # geodesic paths
        no = cz.shape[1]
        q_pca = np.ndarray(shape=(M, Nstd, no), dtype=float)
        f_pca = np.ndarray(shape=(M, Nstd, no), dtype=float)

        for k in range(0, no):
            for l in range(0, Nstd):
                qhat = mqn + np.dot(Psi_q[:, k], stds[l] * np.sqrt(sz[k]))
                hhat = np.dot(Psi_h[:, k], (stds[l] * np.sqrt(sz[k])) / C)
                gamhat = fs.geometry.h_to_gam(hhat)

                if srsf:
                    fhat = fs.utility_functions.cumtrapzmid(
                        time,
                        qhat[0:M] * np.fabs(qhat[0:M]),
                        np.sign(qhat[M]) * (qhat[M] * qhat[M]),
                        mididx,
                    )
                    f_pca[:, l, k] = fs.utility_functions.warp_f_gamma(
                        np.linspace(0, 1, M), fhat, gamhat
                    )
                    q_pca[:, l, k] = fs.utility_functions.warp_q_gamma(
                        np.linspace(0, 1, M), qhat[0:M], gamhat
                    )
                else:
                    f_pca[:, l, k] = fs.utility_functions.warp_f_gamma(
                        np.linspace(0, 1, M), qhat, gamhat
                    )
                    q_pca[:, l, k] = fs.f_to_srsf(f_pca[:, l, k],
                                                  np.linspace(0, 1, M))

        self.q_pca = q_pca
        self.f_pca = f_pca
        self.eigs = sz
        self.latent = sz[0:no]
        self.coef = cz[:, 0:no]
        self.U_q = Psi_q
        self.U_h = Psi_h
        self.id = mididx
        self.C = C
        self.h = h
        self.qn1 = qn2
        self.time = time
        self.no = no
        self.stds = stds
        self.mqn = mqn
        self.U = U
        self.U1 = Uh
        self.Uz = Uz
        self.mh = mh

        return

    def project(self, f):
        """
        project new data onto fPCA basis

        Usage: obj.project(f)

        :param f: numpy array (MxN) of N functions on M time

        """

        q1 = fs.f_to_srsf(f, self.time)
        M = self.time.shape[0]
        n = q1.shape[1]
        mq = self.warp_data.mqn
        fn = np.zeros((M, n))
        qn = np.zeros((M, n))
        gam = np.zeros((M, n))
        for ii in range(0, n):
            gam[:, ii] = fs.optimum_reparam(mq, self.time, q1[:, ii])
            fn[:, ii] = fs.warp_f_gamma(self.time, f[:, ii], gam[:, ii])
            qn[:, ii] = fs.f_to_srsf(fn[:, ii], self.time)

        m_new = np.sign(fn[self.id, :]) * np.sqrt(np.abs(fn[self.id, :]))
        qn1 = np.vstack((qn, m_new))
        C = self.C
        
        h = geo.gam_to_h(gam)

        c = (qn1 - self.mqn[:, np.newaxis]).T @ self.U
        ch = (C*h - self.mh[:, np.newaxis]).T @ self.U1

        Xi = np.hstack((c, ch))

        cz = Xi @ self.Uz

        self.new_coef = cz
        self.new_qn1 = qn1
        self.new_h = h

        return

    def plot(self):
        """
        plot plot elastic joint fPCA result

        Usage: obj.plot()
        """
        no = self.no
        M = self.time.shape[0]
        Nstd = self.stds.shape[0]
        num_plot = int(np.ceil(no / 3))
        CBcdict = {
            "Bl": (0, 0, 0),
            "Or": (0.9, 0.6, 0),
            "SB": (0.35, 0.7, 0.9),
            "bG": (0, 0.6, 0.5),
            "Ye": (0.95, 0.9, 0.25),
            "Bu": (0, 0.45, 0.7),
            "Ve": (0.8, 0.4, 0),
            "rP": (0.8, 0.6, 0.7),
        }
        cl = sorted(CBcdict.keys())
        k = 0
        for ii in range(0, num_plot):
            if k > (no - 1):
                break

            fig, ax = plt.subplots(2, 3)

            for k1 in range(0, 3):
                k = k1 + (ii) * 3
                axt = ax[0, k1]
                if k > (no - 1):
                    break

                for l in range(0, Nstd):
                    axt.plot(self.time, self.q_pca[0:M, l, k], color=CBcdict[cl[l]])

                axt.set_title("q domain: PD %d" % (k + 1))

                axt = ax[1, k1]
                for l in range(0, Nstd):
                    axt.plot(self.time, self.f_pca[:, l, k], color=CBcdict[cl[l]])

                axt.set_title("f domain: PD %d" % (k + 1))

            fig.set_tight_layout(True)

        cumm_coef = 100 * np.cumsum(self.latent) / sum(self.latent)
        idx = np.arange(0, self.latent.shape[0]) + 1
        plot.f_plot(idx, cumm_coef, "Coefficient Cumulative Percentage")
        plt.ylabel("Percentage")
        plt.xlabel("Index")
        plt.show()

        return


def jointfPCAd(qn, vec, C, m, mu_psi, parallel, cores):
    (M, N) = qn.shape
    g = np.vstack((qn, C * vec))

    mu_q = qn.mean(axis=1)
    mu_g = g.mean(axis=1)

    K = np.cov(g)
    U, s, V = svd(K)

    a = np.zeros((N, m))
    for i in range(0, N):
        for j in range(0, m):
            tmp = g[:, i] - mu_g
            a[i, j] = np.dot(tmp.T, U[:, j])

    qhat = np.tile(mu_q, (N, 1))
    qhat = qhat.T
    qhat = qhat + np.dot(U[0:M, 0:m], a.T)

    vechat = np.dot(U[M:, 0:m], a.T / C)
    psihat = np.zeros((M - 1, N))
    gamhat = np.zeros((M - 1, N))
    if parallel:
        out = Parallel(n_jobs=cores)(
            delayed(jfpca_sub)(mu_psi, vechat[:, n]) for n in range(N)
        )
        gamhat = np.array(out)
        gamhat = gamhat.transpose()
    else:
        for ii in range(0, N):
            psihat[:, ii] = geo.exp_map(mu_psi, vechat[:, ii])
            gam_tmp = cumulative_trapezoid(
                psihat[:, ii] * psihat[:, ii], np.linspace(0, 1, M - 1), 
                initial=0)
            gamhat[:, ii] = (gam_tmp - gam_tmp.min()) / (gam_tmp.max() - gam_tmp.min())

    U = U[:, 0:m]
    s = s[0:m]

    return qhat, gamhat, a, U, s, mu_g, g, K


def jointfPCAhd(qn, h, C, var_exp=None):
    (M, N) = qn.shape

    # Run Univariate fPCA
    # q space
    K = np.cov(qn)

    mqn = qn.mean(axis=1)

    U, s, V = svd(K)
    U, V = uf.svd_flip(U, V)

    cumm_coef = np.cumsum(s) / s.sum()
    no_q = int(np.argwhere(cumm_coef >= var_exp)[0][0])

    c = (qn - mqn[:, np.newaxis]).T @ U
    c = c[:, 0:no_q]
    U = U[:, 0:no_q]

    # h space
    hc = C * h
    mh = np.mean(hc, axis=1)
    Kh = np.cov(hc)

    Uh, sh, Vh = svd(Kh)
    Uh, Vh = uf.svd_flip(Uh, Vh)

    cumm_coef = np.cumsum(sh) / sh.sum()
    no_h = int(np.argwhere(cumm_coef >= var_exp)[0][0]) + 1

    ch = (hc - mh[:, np.newaxis]).T @ Uh
    ch = ch[:, 0:no_h]
    Uh = Uh[:, 0:no_h]

    # Run Multivariate fPCA
    Xi = np.hstack((c, ch))
    Z = 1 / (Xi.shape[0] - 1) * Xi.T @ Xi

    Uz, sz, Vz = svd(Z)
    Uz, Vz = uf.svd_flip(Uz, Vz)

    cz = Xi @ Uz

    Psi_q = U @ Uz[0:no_q, :]
    Psi_h = Uh @ Uz[no_q:, :]

    hhat = Psi_h @ (cz).T
    gamhat = fs.geometry.h_to_gam(hhat/C)

    qhat = Psi_q @ cz.T + mqn[:, np.newaxis]

    return qhat, gamhat, cz, Psi_q, Psi_h, sz, U, Uh, Uz


def find_C_h(C, qn, h, q0, var_exp, parallel, cores, srsf):
    qhat, gamhat, cz, Psi_q, Psi_h, sz, U, Uh, Uz = jointfPCAhd(qn, h, C, var_exp)
    (M, N) = qn.shape
    if srsf:
        time = np.linspace(0, 1, M - 1)
    else:
        time = np.linspace(0, 1, M)

    d = np.zeros(N)
    if parallel:
        out = Parallel(n_jobs=cores)(
            delayed(find_C_sub)(time, qhat[:, n], gamhat[:, n], q0[:, n], srsf)
            for n in range(N)
        )
        d = np.array(out)
    else:
        for i in range(0, N):
            if srsf:
                tmp = uf.warp_q_gamma(time, qhat[0: (M - 1), i],
                                      uf.invertGamma(gamhat[:, i]))
            else:
                tmp = uf.warp_f_gamma(time, qhat[:, i],
                                      uf.invertGamma(gamhat[:, i]))
            d[i] = trapezoid((tmp - q0[:, i]) * (tmp - q0[:, i]), time)

    dout = d.mean()

    return dout


def jfpca_sub(mu_psi, vechat):
    M = mu_psi.shape[0]
    psihat = geo.exp_map(mu_psi, vechat)
    gam_tmp = cumulative_trapezoid(psihat * psihat, np.linspace(0, 1, M), 
                                   initial=0)
    gamhat = (gam_tmp - gam_tmp.min()) / (gam_tmp.max() - gam_tmp.min())

    return gamhat


def find_C(C, qn, vec, q0, m, mu_psi, parallel, cores):
    qhat, gamhat, a, U, s, mu_g, g, K = jointfPCAd(
        qn, vec, C, m, mu_psi, parallel, cores
    )
    (M, N) = qn.shape
    time = np.linspace(0, 1, M - 1)

    d = np.zeros(N)
    if parallel:
        out = Parallel(n_jobs=cores)(
            delayed(find_C_sub)(time, qhat[:, n], gamhat[:, n], 
                                q0[:, n])
            for n in range(N)
        )
        d = np.array(out)
    else:
        for i in range(0, N):
            tmp = uf.warp_q_gamma(
                time, qhat[0: (M - 1), i], uf.invertGamma(gamhat[:, i])
            )
            d[i] = trapezoid((tmp - q0[:, i]) * (tmp - q0[:, i]), time)

    out = d.mean()

    return out


def find_C_sub(time, qhat, gamhat, q0, srsf=True):
    if srsf:
        M = qhat.shape[0]
        tmp = uf.warp_q_gamma(time, qhat[0: (M - 1)], uf.invertGamma(gamhat))
    else:
        tmp = uf.warp_f_gamma(time, qhat, uf.invertGamma(gamhat))
    d = trapezoid((tmp - q0) * (tmp - q0), time)

    return d
